/* 
 * tclFileName.c --
 *
 *	This file contains routines for converting file names betwen
 *	native and network form.
 *
 * Copyright (c) 1995 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

static char sccsid[] = "@(#) tclFileName.c 1.10 96/01/24 18:53:05";

#include "tclInt.h"
#include "tclPort.h"
#include "tclRegexp.h"

/*
 * The following regular expression matches the root directory of
 * a network name.  It is rather tricky because it is dependent on
 * the order of the parts in the alternation subexpression.
 * This expression should match strings of the form "//.../.../" or 
 * "//.../...", where the ellipses are strings containing no unquoted
 * slashes.  The first subexpression can be used to extract the root
 * directory without any trailing backslashes.
 */

#define ROOT_PATTERN "^(//(\\\\/|[^/])+/(\\\\/|[^/])+)(/?)"

/*
 * The following regular expression matches DOS drive relative
 * paths in a Network name.  Note that we need the ./ form for relative
 * paths in order to ensure that the path conforms to the //host/root/
 * form of network path.  Otherwise, the first path component after the
 * drive would get treated as part of the mount point.
 */

#define DRIVE_REL_PATTERN "^//\\?/[a-zA-Z]\\\\:[.]?/"

/*
 * The following variables are used to hold precompiled regular expressions
 * for use in filename matching.
 */

static regexp *rootPatternPtr = NULL;
static regexp *driveRelPatternPtr = NULL;

/*
 * The following array contains the path separator characters for
 * each of the defined platforms.  The first character in each string
 * should be the preferred separator when constructing new paths.
 */

static char *separators[] = { "", "/", ":", "\\/:" };

/*
 * The following variable is set in the TclPlatformInit call to one
 * of: TCL_PLATFORM_UNIX, TCL_PLATFORM_MAC, or TCL_PLATFORM_WINDOWS.
 */

TclPlatformType tclPlatform = TCL_PLATFORM_UNKNOWN;

/*
 * Prototypes for local procedures defined in this file:
 */

static char *		DoTildeSubst _ANSI_ARGS_((Tcl_Interp *interp,
			    char *user, Tcl_DString *resultPtr));
static int 		IsNetworkName _ANSI_ARGS_((char *name));
static int		SkipToChar _ANSI_ARGS_((char **stringPtr,
			    char *match));

/*
 *----------------------------------------------------------------------
 *
 * IsNetworkName --
 *
 *	This function checks to see if a file name conforms to the
 *	network name format.
 *
 * Results:
 *	Returns 1 if the file name is in network form, otherwise it
 *	returns 0.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
IsNetworkName(name)
    char *name;			/* File name to be checked. */
{
    char *p;

    /*
     * A filename is a valid network name iff it meets the following
     * conditions:
     *
     * The ':', and '\', characters may not appear unquoted.
     * The quote character '\' may only be used to quote ':', '/', and '\'.
     * The string "//" may only appear at the beginning of a name.
     * The first character may not be a '\'.
     */

    if (*name == '\\') {
	return 0;
    }
    for (p = name; *p != '\0'; p++) {
	if (*p == ':') {
	    return 0;
	} else if (*p == '\\') {
	    p++;
	    if ((*p != '\\') && (*p != '/') && (*p != ':')) {
		return 0;
	    }
	} else if ((*p == '/') && (p[1] == '/') && (p != name)) {
	    return 0;
	}
    }
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * TclConvertToNetwork --
 *
 *	Given a name in native form, produce a name in network form.
 *
 * Results:
 *
 *	The result is a pointer to a static string containing the new
 *	name.  The result may be stored in bufferPtr; the caller must
 *	call Tcl_DStringFree(bufferPtr) to free the name.
 *
 * Side effects:
 *	Information may be left in bufferPtr.
 *
 *----------------------------------------------------------------------
 */

char *
TclConvertToNetwork(interp, name, bufferPtr)
    Tcl_Interp *interp;		/* Interpreter in which to store error
				 * message (if necessary). Ignored
				 * if interp is NULL. */
    char *name;			/* File name in native form. */
    Tcl_DString *bufferPtr;	/* May be used to hold result.  Must not hold
				 * anything at the time of the call, and need
				 * not even be initialized. */
{
    char *p, *string;
    int length;

    Tcl_DStringInit(bufferPtr);

    /*
     * We need to preserve network names across multiple invocations.
     */

    if (IsNetworkName(name)) {
	return name;
    }

    switch (tclPlatform) {
	case TCL_PLATFORM_UNIX:

	    /*
	     * First, we ensure that the result does not start with '\'.
	     */

	    if ((*name == '\\') || (*name == ':')) {
		Tcl_DStringAppend(bufferPtr, "./", 2);
	    }

	    /*
	     * Next, we quote ':' and '\' characters, and eliminate any
	     * redundant '/' characters.
	     */

	    string = name;
	    for (p = name; *p != '\0'; p++) {
		if (*p == '\\' || *p == ':') {
		    Tcl_DStringAppend(bufferPtr, string, p - string);
		    string = p+1;
		    Tcl_DStringAppend(bufferPtr, "\\", 1);
		    Tcl_DStringAppend(bufferPtr, p, 1);
		} else if ((*p == '/') && (p[1] == '/')) {
		    Tcl_DStringAppend(bufferPtr, string, p - string);
		    string = p+1;
		}
		    
	    }
	    Tcl_DStringAppend(bufferPtr, string, -1);
	    return bufferPtr->string;

	case TCL_PLATFORM_MAC:

	    /*
	     * First determine whether the path is relative or absolute.
	     */

	    p = strchr(name, ':');
	    if (p == name) {
		if (p[1] == '\0') {
		    Tcl_DStringAppend(bufferPtr, ".", 1);
		    return bufferPtr->string;
		} else if ((p[1] == '\\') || (p[1] == '/')) {
		    Tcl_DStringAppend(bufferPtr, "./", 2);
		}
	    } else if (p != NULL) {
		Tcl_DStringAppend(bufferPtr, "/", 1);
	    } else if ((*name == '\\') || (*name == '/')) {
		Tcl_DStringAppend(bufferPtr, "./", 2);
	    }

	    string = name;
	    for (p = name; *p != '\0'; p++) {
		if ((*p == '\\') || (*p == '/')) {
		    Tcl_DStringAppend(bufferPtr, string, p - string);
		    string = p+1;
		    Tcl_DStringAppend(bufferPtr, "\\", 1);
		    Tcl_DStringAppend(bufferPtr, p, 1);
		} else if (*p == ':') {
		    Tcl_DStringAppend(bufferPtr, string, p - string);
		    if (p != name) {
			Tcl_DStringAppend(bufferPtr, "/", 1);
		    }
		    while (p[1] == ':') {
			Tcl_DStringAppend(bufferPtr, "../", 3);
			p++;
		    }
		    string = p+1;
		}
	    }
	    Tcl_DStringAppend(bufferPtr, string, -1);
	    return bufferPtr->string;

	case TCL_PLATFORM_WINDOWS:
	    length = strlen(name);
	    if ((length > 1) && (name[1] == ':')) {
		Tcl_DStringAppend(bufferPtr, "//?/x\\:", 7);
		bufferPtr->string[4] = name[0];
		if ((length == 2) 
			|| ((name[2] != '/') && (name[2] != '\\'))) {
		    Tcl_DStringAppend(bufferPtr, "./", 2);
		}
		length = Tcl_DStringLength(bufferPtr);
		Tcl_DStringAppend(bufferPtr, &name[2], -1);
		string = bufferPtr->string + length;
	    } else {
		Tcl_DStringAppend(bufferPtr, name, -1);
		string = bufferPtr->string;
	    }
	    for (p = string; *p != '\0'; p++) {
		if ((*p == '\\') || (*p == '/')) {
		    if ((p != bufferPtr->string)
			    && ((p[1] == '\\') || (p[1] == '/'))) {
			continue;
		    } else {
			*string = '/';
		    }
		} else if (*p == ':') {
		    goto error;
		} else {
		    *string = *p;
		}
		string++;
	    }
	    *string = '\0';
	    Tcl_DStringSetLength(bufferPtr, (int) strlen(bufferPtr->string));
	    return bufferPtr->string;

	case TCL_PLATFORM_UNKNOWN:
	    goto error;
    }
    error:
    if (interp != NULL) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "couldn't convert \"", name,
		"\" to network form", (char *) NULL);
    }
    Tcl_DStringFree(bufferPtr);
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * TclConvertToNative --
 *
 *	Given a name in network form, produce a name in native form.
 *
 * Results:
 *	The result is a pointer to a static string containing the new
 *	name.  The result may be stored in bufferPtr; the caller must
 *	call Tcl_DStringFree(bufferPtr) to free the name.
 *
 * Side effects:
 *	Information may be left in bufferPtr.
 *
 *----------------------------------------------------------------------
 */

char *
TclConvertToNative(interp, name, bufferPtr)
    Tcl_Interp *interp;		/* Interpreter in which to store error
				 * message (if necessary).  Ignored
				 * if interp is NULL. */
    char *name;			/* File name in network form. */
    Tcl_DString *bufferPtr;	/* May be used to hold result.  Must not hold
				 * anything at the time of the call, and need
				 * not even be initialized. */
{
    char *p, *string;

    Tcl_DStringInit(bufferPtr);

    if (*name == '\0') {
	return bufferPtr->string;
    }

    if (!IsNetworkName(name)) {
	return name;
    }

    switch (tclPlatform) {
	case TCL_PLATFORM_UNIX:

	    /*
	     * Disallow names beginning with //
	     */

	    if ((name[0] == '/') && (name[1] == '/')) {
		goto error;
	    }

	    /*
	     * Unquote quoted characters, failing if we find a quoted
	     * forward slash.
	     */

	    Tcl_DStringAppend(bufferPtr, name, -1);
	    string = bufferPtr->string;
	    for (p = string; *p != '\0'; string++, p++) {
		if (*p == '\\') {
		    p++;
		    if (*p == '/') {
			goto error;
		    }
		}
		*string = *p;
	    }
	    *string = '\0';
	    Tcl_DStringSetLength(bufferPtr, (int) strlen(bufferPtr->string));
	    return bufferPtr->string;

	case TCL_PLATFORM_MAC: {
	    int absolute, inWord;

	    /*
	     * Disallow names beginning with //
	     */

	    if ((name[0] == '/') && (name[1] == '/')) {
		goto error;
	    }

	    absolute = (name[0] == '/');

	    /*
	     * Check for relative or absolute paths and set up the
	     * destination buffer accordingly.
	     */

	    if (absolute) {
		Tcl_DStringAppend(bufferPtr, name+1, -1);
		string = bufferPtr->string;
	    } else {
		Tcl_DStringAppend(bufferPtr, ":", 1);
		Tcl_DStringAppend(bufferPtr, name, -1);
		string = bufferPtr->string+1;
	    }

	    /*
	     * Unquote quoted characters, failing if we find a quoted
	     * colon.  Convert / to :, ../ to ::, and dropping ./.
	     */

	    p = string;
	    inWord = 0;
	    while (*p != '\0') {
		if (*p == '\\') {
		    p++;
		    if (*p == ':') {
			goto error;
		    }
		    *string++ = *p++;
		    inWord = 1;
		} else if (*p == '/') {
		    p++;
		    *string++ = ':';
		    inWord = 0;
		} else if (!inWord && (*p == '.')) {
		    p++;
		    if (*p == '/') {
			p++;
		    } else if ((*p == '.')
			    && ((p[1] == '/') || (p[1] == '\0'))) {
			*string++ = ':';
		    } else if (*p != '\0') {
			*string++ = '.';
			inWord = 1;
		    }
		} else {
		    *string++ = *p++;
		    inWord = 1;
		}
	    }
	    *string = '\0';
	    Tcl_DStringSetLength(bufferPtr, (int) strlen(bufferPtr->string));
	    if (absolute) {
		if (Tcl_DStringLength(bufferPtr) == 0) {
		    goto error;
		}
		if (strchr(bufferPtr->string, ':') == NULL) {
		    Tcl_DStringAppend(bufferPtr, ":", 1);
		}
	    }
	    return bufferPtr->string;
	}

	case TCL_PLATFORM_WINDOWS:

	    /*
	     * Check for network form of DOS drive relative paths.
	     */

	    if (driveRelPatternPtr == NULL) {
		driveRelPatternPtr = TclRegComp(DRIVE_REL_PATTERN);
	    }
	    
	    if (TclRegExec(driveRelPatternPtr, name, name) == 1) {
		/*
		 * Extract the drive letter and strip off any leading ./
		 */

		Tcl_DStringAppend(bufferPtr, "?:", 2);
		bufferPtr->string[0] = name[4];
		if (name[7] == '.') {
		    string = &name[9];
		} else {
		    string = &name[7];
		}
	    } else {
		string = name;
	    }

	    /*
	     * Now change '/' into '\', and make sure there are no quoted
	     * characters.
	     */

	    Tcl_DStringAppend(bufferPtr, string, -1);
	    for (p = bufferPtr->string; *p != '\0'; p++) {
		if (*p == '\\') {
		    goto error;
		} else if (*p == '/') {
		    *p = '\\';
		}
	    }
	    return bufferPtr->string;

	case TCL_PLATFORM_UNKNOWN:
	    goto error;
    }

    error:
    if (interp != NULL) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "couldn't convert \"", name,
		"\" to native form", (char *) NULL);
    }
    Tcl_DStringFree(bufferPtr);
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * TclParseFileName --
 *
 *	This function determines the end of the root directory prefix,
 *	the last path separator character, and the beginning of
 *	the extension part of the file name.  If the root directory
 *	prefix is not / or \, then any trailing / or \ characters
 *	will be dropped from the result.
 *
 * Results:
 *	Returns a pointer to the end of the root directory specifier,
 *	the last separator, and the beginning of the extension.  If
 *	a given part is not present, then it returns NULL for that part.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
TclParseFileName(name, rootEndPtr, lastSeparatorPtr, extensionPtr)
    char *name;			/* File name to parse. */
    char **rootEndPtr;		/* Out parameter, points to end of root or
				 * NULL if none. */
    char **lastSeparatorPtr;	/* Out parameter, points to last separator or
				 * NULL if none. */
    char **extensionPtr;	/* Out parameter, points to last '.' or
				 * NULL if none. */
{
    char *p;

    /*
     * First find the last character in the root and the last
     * directory separator.
     */

    if (IsNetworkName(name)) {
	if (!rootPatternPtr) {
	    rootPatternPtr = TclRegComp(ROOT_PATTERN);
	}
	if (TclRegExec(rootPatternPtr, name, name) == 1) {
	    p = rootPatternPtr->endp[0];
	    *rootEndPtr = rootPatternPtr->startp[4];
	    *lastSeparatorPtr = *rootEndPtr;
	} else if (*name == '/') {
	    p = name+1;
	    *rootEndPtr = name;
	    *lastSeparatorPtr = name;
	} else {
	    p = name;
	    *rootEndPtr = NULL;
	    *lastSeparatorPtr = NULL;
	}
	
	while (*p != '\0') {
	    if ((*p == '\\') && (p[1] == '/')) {
		p++;
	    } else if (*p == '/') {
		*lastSeparatorPtr = p;
	    }
	    p++;
	}
    } else {
	switch (tclPlatform) {
	    case TCL_PLATFORM_UNIX:
		*rootEndPtr = (*name == '/') ? name : NULL;
		*lastSeparatorPtr = strrchr(name, '/');
		break;

	    case TCL_PLATFORM_MAC:
		*rootEndPtr = (*name == ':') ? NULL : strchr(name, ':');
		*lastSeparatorPtr = strrchr(name, ':');
		break;

	    case TCL_PLATFORM_WINDOWS:
		p = name;
		if ((strlen(name) > (size_t)1) && (p[1] == ':')) {
		    if ((p[2] == '/') || (p[2] == '\\')) {
			*rootEndPtr = &p[2];
		    } else {
			*rootEndPtr = &p[1];
		    }
		    p = (*rootEndPtr) + 1;
		} else {
		    *rootEndPtr = NULL;
		}
		*lastSeparatorPtr = *rootEndPtr;
	
		while (*p != '\0') {
		    if ((*p == '/') || (*p == '\\')) {
			*lastSeparatorPtr = p;
		    }
		    p++;
		}
		break;

	    case TCL_PLATFORM_UNKNOWN:
		*extensionPtr = *rootEndPtr = *lastSeparatorPtr = NULL;
		return;
	}
    }
    *extensionPtr = strrchr(name, '.');
    if ((*extensionPtr != NULL) && (*lastSeparatorPtr != NULL)
	    && (*lastSeparatorPtr > *extensionPtr)) {
	*extensionPtr = NULL;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * DoTildeSubst --
 *
 *	Given a string following a tilde, this routine returns the
 *	corresponding home directory.
 *
 * Results:
 *	The result is a pointer to a static string containing the home
 *	directory in native format.  If there was an error in processing
 *	the substitution, then an error message is left in interp->result
 *	and the return value is NULL.  The result may be stored in
 *	resultPtr; the caller must call Tcl_DStringFree(resultPtr) to
 *	free the name.
 *
 * Side effects:
 *	Information may be left in resultPtr.
 *
 *----------------------------------------------------------------------
 */

static char *
DoTildeSubst(interp, user, resultPtr)
    Tcl_Interp *interp;		/* Interpreter in which to store error
				 * message (if necessary). */
    char *user;			/* Name of user whose home directory should be
				 * substituted, or "" for current user. */
    Tcl_DString *resultPtr;	/* May be used to hold result.  Must not hold
				 * anything at the time of the call, and need
				 * not even be initialized. */
{
    char *dir;

    if (*user == '\0') {

	/*
	 * Get the home directory of the current user.
	 */

	dir = Tcl_GetVar2(interp, "env", "HOME", TCL_GLOBAL_ONLY);
	if (dir == NULL) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "couldn't find HOME environment ",
		    "variable to expand path", (char *) NULL);
	    return NULL;
	}
	Tcl_DStringInit(resultPtr);
	Tcl_DStringAppend(resultPtr, dir, -1);
    } else {
	if (TclGetUserHome(user, resultPtr) == NULL) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "user \"", user, "\" doesn't exist",
		    (char *) NULL);
	    return NULL;
	}
    }
    return resultPtr->string;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_TranslateFileName --
 *
 *	Given a name starting with a tilde, produce a name where
 *	the tilde and following characters have been replaced by
 *	the home directory location for the named user.  Regardless
 *	of whether the given name contains a leading tilde, the
 *	resulting file name will be in native form.
 *
 * Results:
 *	The result is a pointer to a static string containing
 *	the new name.  If there was an error in processing the
 *	tilde, then an error message is left in interp->result
 *	and the return value is NULL.  The result may be stored
 *	in bufferPtr; the caller must call Tcl_DStringFree(bufferPtr)
 *	to free the name.
 *
 * Side effects:
 *	Information may be left in bufferPtr.
 *
 *----------------------------------------------------------------------
 */

char *
Tcl_TranslateFileName(interp, name, bufferPtr)
    Tcl_Interp *interp;		/* Interpreter in which to store error
				 * message (if necessary). */
    char *name;			/* File name, which may begin with "~"
				 * (to indicate current user's home directory)
				 * or "~<user>" (to indicate any user's
				 * home directory). */
    Tcl_DString *bufferPtr;	/* May be used to hold result.  Must not hold
				 * anything at the time of the call, and need
				 * not even be initialized. */
{
    char *dir, *fileName, c;
    int doSubst;
    register char *tail;
    Tcl_DString temp;
    int tailSep, headSep;


    doSubst = (name[0] == '~');
    fileName = TclConvertToNative(interp, name, bufferPtr);
    if ((fileName == NULL) || (*fileName == '\0')) {
	return fileName;
    } else if (doSubst) {

	/*
	 * Skip to the beginning of the user name, leave tail pointing just
	 * after the username.
	 */

	if ((tclPlatform == TCL_PLATFORM_MAC) && (fileName[0] == ':')) {
	    fileName += 2;
	} else {
	    fileName++;
	}
	tail = strpbrk(fileName, separators[tclPlatform]);
	if (tail == NULL) {
	    tail = fileName + strlen(fileName);
	}

	/*
	 * Perform the tilde substitution.
	 */

	c = *tail;
	*tail = '\0';
	dir = DoTildeSubst(interp, fileName, &temp);
	*tail = c;

	if (dir == NULL) {
	    Tcl_DStringFree(bufferPtr);
	    return NULL;
	}

	/*
	 * Check for double or missing directory separators.
	 */

	if (*tail != '\0') {
	    headSep = (strchr(separators[tclPlatform],
		    temp.string[Tcl_DStringLength(&temp)-1]) != NULL);
	    tailSep = (strchr(separators[tclPlatform], *tail) != NULL);
	    if (headSep && tailSep) {
		tail++;
	    } else if (!headSep && !tailSep) {
		Tcl_DStringAppend(&temp, separators[tclPlatform], 1);
	    }
	}

	/*
	 * Append the rest of the path and copy the result into the
	 * bufferPtr DString.
	 */

	Tcl_DStringAppend(&temp, tail, -1);
	Tcl_DStringSetLength(bufferPtr, (int) 0);
	Tcl_DStringAppend(bufferPtr, temp.string, Tcl_DStringLength(&temp));
	Tcl_DStringFree(&temp);
	fileName = Tcl_DStringValue(bufferPtr);
    }

    /*
     * Ensure that we only use \ for Windows, since some system interfaces
     * actually care.  Also, eliminate any double \ sequences that aren't
     * at the beginning of the path, since this causes major grief on
     * Windows '95 when it tries to interpret the result as a UNC name.
     */

    if (tclPlatform == TCL_PLATFORM_WINDOWS) {
	char *src, *dest, lastChar;

	/*
	 * Handle the first character specially so we don't swallow a
	 * leading // or \\ from a legitimate UNC file name.
	 */

	src = fileName;
	if ((*src == '/') || (*src == '\\')) {
	    *src++ = '\\';
	}
	
	for (dest = src, lastChar = '\0'; *src != '\0'; src++) {
	    *dest = (*src == '/') ? '\\' : *src;
	    if ((*dest == '\\') && (lastChar == '\\')) {
		continue;
	    }
	    lastChar = *dest++;
	}
	*dest = '\0';
    }
	
    return fileName;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_GlobCmd --
 *
 *	This procedure is invoked to process the "glob" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_GlobCmd(dummy, interp, argc, argv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    int i, result, noComplain, firstArg;
    char *thisName, *tail;
    char c;
    Tcl_DString buffer;

    noComplain = 0;
    for (firstArg = 1; (firstArg < argc) && (argv[firstArg][0] == '-');
	    firstArg++) {
	if (strcmp(argv[firstArg], "-nocomplain") == 0) {
	    noComplain = 1;
	} else if (strcmp(argv[firstArg], "--") == 0) {
	    firstArg++;
	    break;
	} else {
	    Tcl_AppendResult(interp, "bad switch \"", argv[firstArg],
		    "\": must be -nocomplain or --", (char *) NULL);
	    return TCL_ERROR;
	}
    }
    if (firstArg >= argc) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" ?switches? name ?name ...?\"", (char *) NULL);
	return TCL_ERROR;
    }

    Tcl_DStringInit(&buffer);
    for (i = firstArg; i < argc; i++) {
	thisName = argv[i];
	Tcl_DStringSetLength(&buffer, (int) 0);

	/*
	 * Perform tilde substitution, if needed.
	 */

	if (thisName[0] == '~') {
	    Tcl_DString temp;
	    char *p, *string;

	    /*
	     * Find the first path separator after the tilde.
	     */

	    for (tail = thisName; *tail != '\0'; tail++) {
		if (*tail == '\\') {
		    if (strchr(separators[tclPlatform], tail[1]) != NULL) {
			break;
		    }
		} else if (strchr(separators[tclPlatform], *tail) != NULL) {
		    break;
		}
	    }

	    /*
	     * Determine the home directory for the specified user.  Note that
	     * we don't allow special characters in the user name.
	     */

	    c = *tail;
	    *tail = '\0';
	    p = strpbrk(thisName+1, "\\[]*?{}");
	    if (p == NULL) {
		thisName = DoTildeSubst(interp, thisName+1, &temp);
	    } else {
		if (!noComplain) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "globbing characters not ",
			    "supported in user names", (char *) NULL);
		}
		thisName = NULL;
	    }
	    *tail = c;
	    if (thisName == NULL) {
		if (noComplain) {
		    Tcl_ResetResult(interp);
		    continue;
		} else {
		    Tcl_DStringFree(&buffer);
		    return TCL_ERROR;
		}
	    }

	    /*
	     * Quote special characters in the substitution result.
	     */

	    string = thisName;
	    for (;;) {
		p = strpbrk(string, "\\[]*?{}");
		if (p == NULL) {
		    Tcl_DStringAppend(&buffer, string, -1);
		    break;
		}
		Tcl_DStringAppend(&buffer, string, p - string);
		Tcl_DStringAppend(&buffer, "\\", 1);
		Tcl_DStringAppend(&buffer, p, 1);
		string = p+1;
	    }
	    Tcl_DStringFree(&temp);

	} else {
	    tail = argv[i];
	}

	/*
	 * Append the remainder of the pattern.
	 */

	Tcl_DStringAppend(&buffer, tail, -1);

	result = TclDoGlob(interp, "", buffer.string);
	if (result != TCL_OK) {
	    if (noComplain) {
		Tcl_ResetResult(interp);
		continue;
	    } else {
		Tcl_DStringFree(&buffer);
		return result;
	    }
	}
    }
    Tcl_DStringFree(&buffer);

    if ((*interp->result == 0) && !noComplain) {
	char *sep = "";

	Tcl_AppendResult(interp, "no files matched glob pattern",
		(argc == 2) ? " \"" : "s \"", (char *) NULL);
	for (i = firstArg; i < argc; i++) {
	    Tcl_AppendResult(interp, sep, argv[i], (char *) NULL);
	    sep = " ";
	}
	Tcl_AppendResult(interp, "\"", (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * SkipToChar --
 *
 *	This function traverses a glob pattern looking for the next
 *	unquoted occurance of the specified character at the same braces
 *	nesting level.
 *
 * Results:
 *	Updates stringPtr to point to the matching character, or to
 *	the end of the string if nothing matched.  The return value
 *	is 1 if a match was found at the top level, otherwise it is 0.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
SkipToChar(stringPtr, match)
    char **stringPtr;			/* Pointer string to check. */
    char *match;			/* Pointer to character to find. */
{
    int quoted, level;
    register char *p;

    quoted = 0;
    level = 0;

    for (p = *stringPtr; *p != '\0'; p++) {
	if (quoted) {
	    quoted = 0;
	    continue;
	}
	if ((level == 0) && (*p == *match)) {
	    *stringPtr = p;
	    return 1;
	}
	if (*p == '{') {
	    level++;
	} else if (*p == '}') {
	    level--;
	} else if (*p == '\\') {
	    quoted = 1;
	}
    }
    *stringPtr = p;
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * TclDoGlob --
 *
 *	This recursive procedure forms the heart of the globbing
 *	code.  It performs a depth-first traversal of the tree
 *	given by the path name to be globbed.  The directory and
 *	remainder are assumed to be native format paths.
 *
 * Results:
 *	The return value is a standard Tcl result indicating whether
 *	an error occurred in globbing.  After a normal return the
 *	result in interp will be set to hold all of the file names
 *	given by the dir and rem arguments.  After an error the
 *	result in interp will hold an error message.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
TclDoGlob(interp, dir, rem)
    Tcl_Interp *interp;			/* Interpreter to use for error
					 * reporting (e.g. unmatched brace). */
    char *dir;				/* Name of a directory at which to
					 * start glob expansion.  This name
					 * is fixed: it doesn't contain any
					 * globbing chars. */
    char *rem;				/* Path to glob-expand. */
{
    int level, baseLength, quoted;
    int result = TCL_OK;
    char *p, *openBrace, *closeBrace, *name, savedChar;
    Tcl_DString newName;

    /*
     * Look for the first matching pair of braces or the first
     * directory separator that is not inside a pair of braces.
     */

    openBrace = closeBrace = NULL;
    level = 0;
    quoted = 0;
    for (p = rem; *p != '\0'; p++) {
	if (quoted) {
	    quoted = 0;
	} else if (*p == '\\') {
	    quoted = 1;
	    if (strchr(separators[tclPlatform], p[1]) != NULL) {
		break;			/* Quoted directory separator. */
	    }
	} else if (strchr(separators[tclPlatform], *p) != NULL) {
	    break;			/* Unquoted directory separator. */
	} else if (*p == '{') {
	    openBrace = p;
	    p++;
	    if (SkipToChar(&p, "}")) {
		closeBrace = p;		/* Balanced braces. */
		break;
	    }
	    Tcl_ResetResult(interp);
	    interp->result = "unmatched open-brace in file name";
	    return TCL_ERROR;
	} else if (*p == '}') {
	    Tcl_ResetResult(interp);
	    interp->result = "unmatched close-brace in file name";
	    return TCL_ERROR;
	}
    }

    Tcl_DStringInit(&newName);

    /*
     * Substitute the alternate patterns from the braces and recurse.
     */

    if (openBrace != NULL) {
	char *element;


	/*
	 * For each element within in the outermost pair of braces,
	 * append the element and the remainder to the fixed portion
	 * before the first brace and recursively call TclDoGlob.
	 */

	Tcl_DStringAppend(&newName, rem, openBrace-rem);
	baseLength = Tcl_DStringLength(&newName);
	*closeBrace = '\0';
	for (p = openBrace; p != closeBrace; ) {
	    p++;
	    element = p;
	    SkipToChar(&p, ",");
	    Tcl_DStringAppend(&newName, element, p-element);
	    Tcl_DStringAppend(&newName, closeBrace+1, -1);
	    result = TclDoGlob(interp, dir, newName.string);
	    if (result != TCL_OK) {
		break;
	    }
	    Tcl_DStringSetLength(&newName, (int) baseLength);
	}
	*closeBrace = '}';
	goto done;
    }

    /*
     * At this point, there are no more brace substitutions to perform on
     * this path component.  The variable p is pointing at a quoted or
     * unquoted directory separator or the end of the string.  So we need
     * to check for special globbing characters and do any required
     * matching.
     */

    Tcl_DStringAppend(&newName, dir, -1);
    baseLength = Tcl_DStringLength(&newName);
    savedChar = *p;
    *p = '\0';
    if (strpbrk(rem, "*[]?\\") != NULL) {
	*p = savedChar;

	/*
	 * Look for matching files in the current directory.  The
	 * implementation of this function is platform specific, but may
	 * recursively call TclDoGlob.  For each file that matches, it will
	 * add the match onto the interp->result, or call TclDoGlob if there
	 * are more characters to be processed.
	 */

	result = TclMatchFiles(interp, &newName, rem, p);
	goto done;
    }
    *p = savedChar;

    /*
     * The current element has no special characters, so we can
     * add it directly to the new name.  If there are more characters
     * to be processed, call TclDoGlob recursively.
     */

    Tcl_DStringAppend(&newName, rem, p-rem);
    if (*p != '\0') {
	if (quoted) {
	    p++;
	}
	if (*p != '\0') {
	    Tcl_DStringAppend(&newName, p, 1);
	    result = TclDoGlob(interp, newName.string, p+1);
	    goto done;
	}
    }
    
    /*
     * There are no more elements in the pattern, so all we need to do
     * is verify the existence of the file and add its name to the list being
     * formed in interp->result.
     */

    name = newName.string;
    if (*name == '\0') {
	name = ".";
    }
    if (access(name, F_OK) == 0) {
	Tcl_AppendElement(interp, name);
    }

    done:
    Tcl_DStringFree(&newName);
    return result;
}
