/* 
 * tclIO.c --
 *
 *	This file provides the generic portions (those that are the same on
 *	all platforms and for all channel types) of Tcl's IO facilities.
 *
 * Copyright (c) 1995 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclIO.c 1.157 96/01/24 18:38:00
 */

#include	"tclInt.h"
#include	"tclPort.h"

/*
 * Make sure that both EAGAIN and EWOULDBLOCK are defined. This does not
 * compile on systems where neither is defined. We want both defined so
 * that we can test safely for both. In the code we still have to test for
 * both because there may be systems on which both are defined and have
 * different values.
 */

#if ((!defined(EWOULDBLOCK)) && (defined(EAGAIN)))
#   define EWOULDBLOCK EAGAIN
#endif
#if ((!defined(EAGAIN)) && (defined(EWOULDBLOCK)))
#   define EAGAIN EWOULDBLOCK
#endif
#if ((!defined(EAGAIN)) && (!defined(EWOULDBLOCK)))
    error one of EWOULDBLOCK or EAGAIN must be defined
#endif

/*
 * struct ChannelBuffer:
 *
 * Buffers data being sent to or from a channel.
 */

typedef struct ChannelBuffer {
    int nextAdded;		/* the next position into which a character
                                 * will be put in the buffer. */
    int nextRemoved;		/* Position of next byte to be removed
                                 * from the buffer. */
    int bufSize;		/* How big is the buffer? */
    struct ChannelBuffer *nextPtr;
    				/* Next buffer in chain. */
    char buf[4];		/* Placeholder for real buffer. The real
                                 * buffer occuppies this space + bufSize-4
                                 * bytes. This must be the last field in
                                 * the structure. */
} ChannelBuffer;

#define CHANNELBUFFER_HEADER_SIZE	(sizeof(ChannelBuffer) - 4)

/*
 * Forward declaration of ChannelHandler; being used in struct TclChannel,
 * below.
 */

typedef struct ChannelHandler *ChannelHandlerPtr;

/*
 * Enum for different end of line translation and recognition modes. This
 * is used in the definition of TclChannel, below:
 */

typedef enum EolTranslation {
    TCL_TRANSLATE_AUTO,			/* Automatic detection mode. */
    TCL_TRANSLATE_CR,			/* Eol == \r. */
    TCL_TRANSLATE_LF,			/* Eol == \n. */
    TCL_TRANSLATE_CRLF			/* Eol == \r\n. */
} EolTranslation;

/*
 * Structure to record a close callback. One such record exists for
 * each close callback registered for a channel.
 */

typedef struct CloseCallback {
    Tcl_CloseProc *proc;		/* The procedure to call. */
    ClientData clientData;		/* Arbitrary one-word data to pass
                                         * to the callback. */
    struct CloseCallback *nextPtr;	/* For chaining close callbacks. */
} CloseCallback;

/*
 * The following structure describes the information saved from a call to
 * "channelevent". This is used later when the event being waited for to
 * invoke the saved script in the interpreter designed in this record.
 */

typedef struct EventScriptRecord {
    Tcl_Channel chan;		/* The channel for which this script is
                                 * registered. This is used only when an
                                 * error occurs during evaluation of the
                                 * script, to delete the handler. */
    char *script;		/* Script to invoke. */
    Tcl_Interp *interp;		/* In what interpreter to invoke script? */
    int mask;			/* Events must overlap current mask for the
                                 * stored script to be invoked. */
    struct EventScriptRecord *nextPtr;
    				/* Next in chain of records. */
} EventScriptRecord;

/*
 * struct TclChannel:
 *
 * One of these structures is allocated for each open channel. It contains data
 * specific to the channel but which belongs to the generic part of the Tcl
 * channel mechanism, and it points at an instance specific (and type specific) * instance data, and at a channel type structure.
 */

typedef struct TclChannel {
    char	*channelName;	/* The name of the channel instance in Tcl
                                 * commands. Storage is owned by the generic IO
                                 * code,  is dynamically allocated. */
    int		flags;		/* ORed combination of the flags defined
                                 * below. */
    EolTranslation translation;	/* What translation to apply for end of line
                                 * sequences? */    
    int		latentError;	/* Non-zero if an error report was deferred
                                 * because it happened in the background. The
                                 * value is the POSIX error code. */
    ClientData instanceData;	/* Instance specific data. */
    Tcl_File inFile;		/* File to use for input, or NULL. */
    Tcl_File outFile;		/* File to use for output, or NULL. */
    Tcl_ChannelType *typePtr;	/* Pointer to channel type structure. */
    int		refCount;	/* How many interpreters hold references to
                                 * this IO channel? */
    CloseCallback *closeCallbacks;
    				/* Callbacks registered to be called when the
                                 * channel is closed. */
    ChannelBuffer *curOutPtr;	/* Current output buffer being filled. */
    ChannelBuffer *outQueueHead;/* Points at first buffer in output queue. */
    ChannelBuffer *outQueueTail;/* Points at last buffer in output queue. */

    ChannelBuffer *inQueueHead;	/* Points at first buffer in input queue. */
    ChannelBuffer *inQueueTail;	/* Points at last buffer in input queue. */

    ChannelHandlerPtr channelHandlers;
    				/* List of channel handlers registered
                                 * for this channel. */
    int interestMask;		/* Mask of all events this channel has
                                 * handlers for. */
    struct TclChannel *nextChanPtr;
    				/* Chain of all channels currently open. */
    EventScriptRecord *scriptRecordPtr;
    				/* Chain of all scripts registered for
                                 * event handlers ("fileevent") on this
                                 * channel. */
} TclChannel;
    
/*
 * Values for the flags field in TclChannel. Any ORed combination of the
 * following flags can be stored in the field. These flags record various
 * options and state bits about the channel. In addition to the flags below,
 * the channel can also have TCL_READABLE (1<<1) and TCL_WRITABLE (1<<2) set.
 */

#define TCL_ASYNC_CHANNEL	(1<<3)	/* Channel is currently in
					 * nonblocking mode. */
#define TCL_LINE_BUFFERED	(1<<4)	/* Output to the channel must be
					 * flushed after every newline. */
#define TCL_BUFFER_READY	(1<<5)	/* Current output buffer (the
					 * curOutPtr field in the
                                         * IOChannel structure) is
					 * ready for output to channel. */
#define TCL_ASYNC_FLUSH_ACTIVE	(1<<6)	/* A background flush of the
					 * queued output buffers has been
                                         * scheduled. */
#define TCL_CHANNEL_CLOSED	(1<<7)	/* Channel has been closed. No
					 * further Tcl-level IO on the
                                         * channel is allowed. */
#define	TCL_CHANNEL_EOF		(1<<8)	/* EOF occurred on this channel.
					 * This bit is cleared before every
                                         * input operation. */
#define TCL_CHANNEL_BLOCKED	(1<<9)	/* EWOULDBLOCK or EAGAIN occurred
					 * on this channel. This bit is
                                         * cleared before every input or
                                         * output operation. */
#define TCL_INPUT_SAW_CR	(1<<10)	/* Channel is in network eol input
					 * translation mode and the last
                                         * byte seen was a "\r". */
#define TCL_INPUT_EOF_PENDING	(1<<11)	/* End of file with a pending "\r".
					 * We already produced the "\r". */

/*
 * For each channel handler registered in a call to Tcl_CreateChannelHandler,
 * there is one record of the following type. All of these records are
 * chained together in a singly linked list.
 */

typedef struct ChannelHandler {
    TclChannel *chanPtr;	/* The channel structure for this channel. */
    int mask;			/* Mask of desired events. */
    Tcl_FileProc *proc;		/* Procedure to call in the type of
                                 * Tcl_CreateFileHandler. */
    ClientData clientData;	/* Argument to pass to procedure. */
    struct ChannelHandler *nextPtr;
    				/* Next one in list of registered handlers. */
} ChannelHandler;

/*
 * List of all channels currently open.
 */

static TclChannel *firstChanPtr = (TclChannel *) NULL;

/*
 * Has the channel event source been created and registered with the
 * notifier?
 */

static int channelEventSourceCreated = 0;

/*
 * The following structure describes the event that is added to the Tcl
 * event queue by the channel handler check procedure.
 */

typedef struct ChannelHandlerEvent {
    Tcl_Event header;		/* Information that is standard for all
                                 * events. */
    TclChannel *chanPtr;	/* The channel that is ready. Used to find
                                 * the ChannelHandler structure for this
                                 * channel. */
    int readyMask;		/* Mask of events that have occurred on
                                 * this channel. */
} ChannelHandlerEvent;

/*
 * Static functions in this file:
 */

static int		ChannelEventsDeleteProc _ANSI_ARGS_((
			    Tcl_Event *evPtr, ClientData clientData));
static void		ChannelEventDispatcher _ANSI_ARGS_((
			    ClientData clientData, int mask));
static int		ChannelHandlerEventProc _ANSI_ARGS_((
			    Tcl_Event *evPtr, int flags));
static void		ChannelHandlerCheckProc _ANSI_ARGS_((
			    ClientData clientData, int flags));
static void		ChannelHandlerExitProc _ANSI_ARGS_((ClientData data));
static void		ChannelHandlerSetupProc _ANSI_ARGS_((
			    ClientData clientData, int flags));
static void		ChannelEventScriptInvoker _ANSI_ARGS_((
			    ClientData clientData, int flags));
static void		CloseChannelsOnExit _ANSI_ARGS_((ClientData data));
static int		CopyAndTranslateInput _ANSI_ARGS_((
			    TclChannel *chanPtr, char *result, int space));
static void		CreateScriptRecord _ANSI_ARGS_((
			    Tcl_Interp *interp, TclChannel *chanPtr,
                            int mask, char *script));
static void		DeleteChannelTable _ANSI_ARGS_((
			    ClientData clientData, Tcl_Interp *interp));
static void		DeleteScriptRecord _ANSI_ARGS_((Tcl_Interp *interp,
        		    TclChannel *chanPtr, int mask));
static void		DiscardInputQueued _ANSI_ARGS_((
			    TclChannel *chanPtr));
static int		FlushChannel _ANSI_ARGS_((TclChannel *chanPtr,
			    int calledFromAsyncFlush));
static int		GetInputBuffer _ANSI_ARGS_((TclChannel *chanPtr,
        		    int toRead));
static int		GetNewline _ANSI_ARGS_((TclChannel *chanPtr));
static void		InitializeChannelTable _ANSI_ARGS_((
			    Tcl_Interp *interp));
static void		ReturnScriptRecord _ANSI_ARGS_((Tcl_Interp *interp,
		            TclChannel *chanPtr, int mask));

/*
 *----------------------------------------------------------------------
 *
 * Tcl_CreateCloseHandler
 *
 *	Creates a close callback which will be called when the channel is
 *	closed.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Causes the callback to be called in the future when the channel
 *	will be closed.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_CreateCloseHandler(chan, proc, clientData)
    Tcl_Channel chan;		/* The channel for which to create the
                                 * close callback. */
    Tcl_CloseProc *proc;	/* The callback routine to call when the
                                 * channel will be closed. */
    ClientData clientData;	/* Arbitrary data to pass to the
                                 * close callback. */
{
    TclChannel *chanPtr;
    CloseCallback *cbPtr;

    chanPtr = (TclChannel *) chan;

    cbPtr = (CloseCallback *) ckalloc((unsigned) sizeof(CloseCallback));
    cbPtr->proc = proc;
    cbPtr->clientData = clientData;

    cbPtr->nextPtr = chanPtr->closeCallbacks;
    chanPtr->closeCallbacks = cbPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_DeleteCloseHandler --
 *
 *	Removes a callback which would have been called on closing
 *	the channel.
 *
 * Results:
 *	The callback data that was registered with the callback.
 *
 * Side effects:
 *	The callback will not be called in the future when the channel
 *	is eventually closed.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_DeleteCloseHandler(chan, proc, clientData)
    Tcl_Channel chan;		/* The channel for which to cancel the
                                 * close callback. */
    Tcl_CloseProc *proc;	/* The procedure for the callback to
                                 * remove. */
    ClientData clientData;	/* The callback data for the callback
                                 * to remove. */
{
    TclChannel *chanPtr;
    CloseCallback *cbPtr, *cbPrevPtr;

    chanPtr = (TclChannel *) chan;
    for (cbPtr = chanPtr->closeCallbacks, cbPrevPtr = (CloseCallback *) NULL;
             cbPtr != (CloseCallback *) NULL;
             cbPtr = cbPtr->nextPtr) {
        if ((cbPtr->proc == proc) && (cbPtr->clientData == clientData)) {
            if (cbPrevPtr == (CloseCallback *) NULL) {
                chanPtr->closeCallbacks = cbPtr->nextPtr;
            } else {
                cbPrevPtr = cbPtr->nextPtr;
            }
            ckfree((char *) cbPtr);
            break;
        }
    }
}

/*
 *----------------------------------------------------------------------
 *
 * CloseChannelsOnExit --
 *
 *	Closes all the existing channels, on exit. This	routine is called
 *	during exit processing.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Closes all channels.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
CloseChannelsOnExit(clientData)
    ClientData clientData;		/* NULL - unused. */
{
    TclChannel *chanPtr;		/* Iterates over open channels. */
    TclChannel *nextChanPtr;		/* Iterates over open channels. */


    for (chanPtr = firstChanPtr; chanPtr != (TclChannel *) NULL;
             chanPtr = nextChanPtr) {
        nextChanPtr = chanPtr->nextChanPtr;
                                 
        /*
         * Switch the channel back into synchronous mode to ensure that it
         * gets flushed now.
         */

        (void) Tcl_SetChannelOption(NULL, (Tcl_Channel) chanPtr, "-blocking",
                "on");

        /*
         * Close it only if the refcount indicates that the channel is not
         * referenced from any interpreter. If it is, that interpreter will
         * close the channel when it gets destroyed.
         */
    
        if (chanPtr->refCount <= 0) {
            Tcl_Close((Tcl_Channel) chanPtr);
        }
    }
}

/*
 *----------------------------------------------------------------------
 *
 * InitializeChannelTable --
 *
 *	Initializes the channel table for an interpreter. Also inserts
 *	channels for stdin, stdout and stderr if the interpreter is not
 *	safe.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Initializes the channel table for an interpreter. May create
 *	channels for stdin, stdout and stderr.
 *
 *----------------------------------------------------------------------
 */

static void
InitializeChannelTable(interp)
    Tcl_Interp *interp;
{
    Tcl_HashTable *hTblPtr;	/* Hash table of channels. */
    Tcl_Channel stdinChannel, stdoutChannel, stderrChannel;

    hTblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp, "tclIO", NULL);
    if (hTblPtr == (Tcl_HashTable *) NULL) {
        hTblPtr = (Tcl_HashTable *) ckalloc((unsigned) sizeof(Tcl_HashTable));
        Tcl_InitHashTable(hTblPtr, TCL_STRING_KEYS);

        (void) Tcl_SetAssocData(interp, "tclIO",
                (Tcl_InterpDeleteProc *) DeleteChannelTable,
                (ClientData) hTblPtr);

        /*
         * If the interpreter is trusted (not "safe"), insert channels
         * for stdin, stdout and stderr (possibly creating them in the
         * process).
         */

        if (Tcl_IsSafe(interp) == 0) {
            TclGetStdChannels(&stdinChannel, &stdoutChannel, &stderrChannel);
            if (stdinChannel != NULL) {
                Tcl_RegisterChannel(interp, stdinChannel);
            }
            if (stdoutChannel != NULL) {
                Tcl_RegisterChannel(interp, stdoutChannel);
            }
            if (stderrChannel != NULL) {
                Tcl_RegisterChannel(interp, stderrChannel);
            }
        }

    }
}

/*
 *----------------------------------------------------------------------
 *
 * DeleteChannelTable --
 *
 *	Deletes the channel table for an interpreter, closing any open
 *	channels whose refcount reaches zero.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Deletes the hash table of channels. May close channels. May flush
 *	output on closed channels. Removes any channeEvent handlers that were
 *	registered in this interpreter.
 *
 *----------------------------------------------------------------------
 */

static void
DeleteChannelTable(clientData, interp)
    ClientData clientData;	/* The per-interpreter data structure. */
    Tcl_Interp *interp;		/* The interpreter being deleted. */
{
    Tcl_HashTable *hTblPtr;	/* The hash table. */
    Tcl_HashSearch hSearch;	/* Search variable. */
    Tcl_HashEntry *hPtr;	/* Search variable. */
    TclChannel *chanPtr;	/* Channel being deleted. */
    EventScriptRecord *sPtr, *prevPtr, *nextPtr;
    				/* Variables to loop over all channel events
                                 * registered, to delete the ones that refer
                                 * to the interpreter being deleted. */

    /*
     * Delete all the registered channels - this will close channels whose
     * refcount reaches zero.
     */
    
    hTblPtr = (Tcl_HashTable *) clientData;
    for (hPtr = Tcl_FirstHashEntry(hTblPtr, &hSearch);
             hPtr != (Tcl_HashEntry *) NULL;
             hPtr = Tcl_FirstHashEntry(hTblPtr, &hSearch)) {

        chanPtr = (TclChannel *) Tcl_GetHashValue(hPtr);

        /*
         * Remove any fileevents registered in this interpreter.
         */
        
        for (sPtr = chanPtr->scriptRecordPtr,
                 prevPtr = (EventScriptRecord *) NULL;
                 sPtr != (EventScriptRecord *) NULL;
                 sPtr = nextPtr) {
            nextPtr = sPtr->nextPtr;
            if (sPtr->interp == interp) {
                if (prevPtr == (EventScriptRecord *) NULL) {
                    chanPtr->scriptRecordPtr = nextPtr;
                } else {
                    prevPtr->nextPtr = nextPtr;
                }

                Tcl_DeleteChannelHandler((Tcl_Channel) chanPtr,
                        ChannelEventScriptInvoker, (ClientData) sPtr);

                ckfree(sPtr->script);
                ckfree((char *) sPtr);
                
            } else {
                prevPtr = sPtr;
            }
        }
        Tcl_UnregisterChannel(interp, (Tcl_Channel) chanPtr);
    }
    Tcl_DeleteHashTable(hTblPtr);
    ckfree((char *) hTblPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_UnregisterChannel --
 *
 *	Deletes the hash entry for a channel associated with an interpreter.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Deletes the hash entry for a channel associated with an interpreter.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_UnregisterChannel(interp, chan)
    Tcl_Interp *interp;		/* Interpreter in which channel is defined. */
    Tcl_Channel chan;		/* Channel to delete. */
{
    Tcl_HashTable *hTblPtr;	/* Hash table of channels. */
    Tcl_HashEntry *hPtr;	/* Search variable. */
    TclChannel *chanPtr;	/* The real IO channel. */
    int result;			/* Of calling TclCleanupCommandChannel. */

    chanPtr = (TclChannel *) chan;
    hTblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp, "tclIO", NULL);
    if (hTblPtr == (Tcl_HashTable *) NULL) {
        return TCL_OK;
    }
    hPtr = Tcl_FindHashEntry(hTblPtr, chanPtr->channelName);
    if (hPtr == (Tcl_HashEntry *) NULL) {
        return TCL_OK;
    }
    if ((TclChannel *) Tcl_GetHashValue(hPtr) != chanPtr) {
        return TCL_OK;
    }
    Tcl_DeleteHashEntry(hPtr);
    chanPtr->refCount--;
    if (chanPtr->refCount <= 0) {

        /*
         * If the channel is not going to be closed asynchronously, then
         * do special cleanup for command channels. This includes waiting
         * for the child processes, detaching them, collecting the error
         * output and closing the error channel. If there is any error
         * information or an error during the wait for the children it
         * will be reported. If the children are detached and the channel
         * is closed asynchronously, no errors will be reported.
         */
        
        if (chanPtr->flags & TCL_ASYNC_FLUSH_ACTIVE) {
            result = TCL_OK;
        } else {
            if ((chanPtr->curOutPtr != (ChannelBuffer *) NULL) &&
                    (chanPtr->curOutPtr->nextAdded > 0)) {
                chanPtr->flags |= TCL_BUFFER_READY;
            }

            /*
             * Any errors during flushing are ignored. Is this
             * safe?
             */
            
            (void) FlushChannel(chanPtr, 0);
            result = TclCleanupCommandChannel(interp, chan, 0);
        }
        
        if (Tcl_Close(chan) != TCL_OK) {
            return TCL_ERROR;
        }
        return result;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_RegisterChannel --
 *
 *	Adds an already-open channel to the channel table of an interpreter.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	May increment the reference count of a channel.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_RegisterChannel(interp, chan)
    Tcl_Interp *interp;		/* Interpreter in which to add the channel. */
    Tcl_Channel chan;		/* The channel to add to this interpreter
                                 * channel table. */
{
    Tcl_HashTable *hTblPtr;	/* Hash table of channels. */
    Tcl_HashEntry *hPtr;	/* Search variable. */
    int new;			/* Is the hash entry new or does it exist? */
    TclChannel *chanPtr;		/* The actual channel. */
    Tcl_Channel foundchan; 	/* Channel already found in interpreter. */

    chanPtr = (TclChannel *) chan;

    InitializeChannelTable(interp);

    if (chanPtr->channelName == (char *) NULL) {
        panic("Tcl_RegisterChannel: channel without name");
    }
    
    foundchan = Tcl_GetChannel(interp, chanPtr->channelName, NULL);
    if (foundchan == chan) {
        return;
    }
    if ((foundchan != (Tcl_Channel) NULL) && (foundchan != chan)) {
        panic("Tcl_RegisterChannel: duplicate channel names");
    }

    Tcl_ResetResult(interp);

    hTblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp, "tclIO", NULL);
    hPtr = Tcl_CreateHashEntry(hTblPtr, chanPtr->channelName, &new);
    if (new == 0) {
        panic("Tcl_RegisterChannel: duplicate channel names");
    }

    Tcl_SetHashValue(hPtr, (ClientData) chanPtr);
    chanPtr->refCount++;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_GetChannel --
 *
 *	Finds an existing Tcl_Channel structure by name in a given
 *	interpreter. This function is public because it is used by
 *	channel-type-specific functions.
 *
 * Results:
 *	A Tcl_Channel or NULL on failure. If failed, interp->result
 *	contains an error message. It also returns, in modePtr, the
 *	modes in which the channel is opened.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_GetChannel(interp, chanName, modePtr)
    Tcl_Interp *interp;		/* Interpreter in which to find or create
                                 * the channel. */
    char *chanName;		/* The name of the channel. */
    int *modePtr;
{
    TclChannel *chanPtr;	/* The actual channel. */
    Tcl_HashTable *hTblPtr;	/* Hash table of channels. */
    Tcl_HashEntry *hPtr;	/* Search variable. */

    InitializeChannelTable(interp);
    hTblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp, "tclIO", NULL);    
    hPtr = Tcl_FindHashEntry(hTblPtr, chanName);
    if (hPtr == (Tcl_HashEntry *) NULL) {
        Tcl_AppendResult(interp, "can not find channel named \"",
                chanName, "\"", (char *) NULL);
        return NULL;
    }

    chanPtr = (TclChannel *) Tcl_GetHashValue(hPtr);
    if (modePtr != NULL) {
        *modePtr = (chanPtr->flags & (TCL_READABLE|TCL_WRITABLE));
    }
    
    return (Tcl_Channel) chanPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_CreateChannel --
 *
 *	Creates a new entry in the hash table for a Tcl_Channel
 *	record.
 *
 * Results:
 *	Returns the new Tcl_Channel.
 *
 * Side effects:
 *	Creates a new Tcl_Channel instance and inserts it into the
 *	hash table.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_CreateChannel(typePtr, chanName, inFile, outFile, instanceData)
    Tcl_ChannelType *typePtr;	/* The channel type record. */
    char *chanName;		/* Name of channel to record. */
    Tcl_File inFile;		/* File to use for input, or NULL. */
    Tcl_File outFile;		/* File to use for output, or NULL. */
    ClientData instanceData;	/* Instance specific data. */
{
    TclChannel *chanPtr;	/* The channel structure newly created. */

    chanPtr = (TclChannel *) ckalloc((unsigned) sizeof(TclChannel));
    
    if (chanName != (char *) NULL) {
        chanPtr->channelName = ckalloc((unsigned) strlen(chanName) + 1);
        strcpy(chanPtr->channelName, chanName);
    } else {
        chanPtr->channelName = (char *) NULL;
    }

    chanPtr->flags = 0;
    if (inFile != (Tcl_File) NULL) {
        chanPtr->flags |= TCL_READABLE;
    }
    if (outFile != (Tcl_File) NULL) {
        chanPtr->flags |= TCL_WRITABLE;
    }

    /*
     * Set the channel up initially in unknown end-of-line translation
     * mode to allow auto-detect of the mode on a channel specific basis.
     */

    chanPtr->translation = TCL_TRANSLATE_AUTO;

    chanPtr->latentError = 0;
    chanPtr->instanceData = instanceData;
    chanPtr->inFile = inFile;
    chanPtr->outFile = outFile;
    chanPtr->typePtr = typePtr;
    chanPtr->refCount = 0;
    chanPtr->closeCallbacks = (CloseCallback *) NULL;
    chanPtr->curOutPtr = (ChannelBuffer *) NULL;
    chanPtr->outQueueHead = (ChannelBuffer *) NULL;
    chanPtr->outQueueTail = (ChannelBuffer *) NULL;
    chanPtr->inQueueHead = (ChannelBuffer *) NULL;
    chanPtr->inQueueTail = (ChannelBuffer *) NULL;
    chanPtr->channelHandlers = (ChannelHandler *) NULL;
    chanPtr->interestMask = 0;

    /*
     * Link the channel into the list of all channels; create an on-exit
     * handler if there is not one already, to close off all the channels
     * in the list on exit.
     */

    chanPtr->nextChanPtr = firstChanPtr;
    if (firstChanPtr == (TclChannel *) NULL) {
        Tcl_CreateExitHandler(CloseChannelsOnExit, (ClientData) NULL);
    }
    firstChanPtr = chanPtr;

    chanPtr->scriptRecordPtr = (EventScriptRecord *) NULL;
    
    return (Tcl_Channel) chanPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_GetChannelName --
 *
 *	Returns the string identifying the channel name. This memory is
 *	owned by the generic layer and should not be modified by the caller.
 *
 * Results:
 *	The string containing the channel name
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

char *
Tcl_GetChannelName(chan)
    Tcl_Channel chan;		/* The channel for which to return the name. */
{
    TclChannel *chanPtr;	/* The actual channel. */

    chanPtr = (TclChannel *) chan;
    return chanPtr->channelName;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_GetChannelType --
 *
 *	Given a channel structure, returns the channel type structure
 *	and its type name.
 *
 * Results:
 *	Returns a pointer to the channel type structure.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Tcl_ChannelType *
Tcl_GetChannelType(chan)
    Tcl_Channel chan;			/* The channel to return type for. */
{
    TclChannel *chanPtr = (TclChannel *) chan;
    return chanPtr->typePtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_GetChannelFile --
 *
 *	Returns the file associated with this channel channel if the channel
 *	is file based.
 *
 * Results:
 *	The file or NULL if failed (e.g. the channel is not open for the
 *	requested direction).
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Tcl_File
Tcl_GetChannelFile(chan, direction)
    Tcl_Channel chan;			/* The channel to get file from. */
    int direction;			/* TCL_WRITABLE or TCL_READABLE. */
{
    TclChannel *chanPtr;		/* The actual channel. */

    chanPtr = (TclChannel *) chan;
    switch (direction) {
        case TCL_WRITABLE:
            return chanPtr->outFile;
        case TCL_READABLE:
            return chanPtr->inFile;
        default:
            return NULL;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_GetChannelInstanceData --
 *
 *	Returns the client data associated with a channel.
 *
 * Results:
 *	The client data.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

ClientData
Tcl_GetChannelInstanceData(chan)
    Tcl_Channel chan;		/* Channel for which to return client data. */
{
    TclChannel *chanPtr;

    chanPtr = (TclChannel *) chan;
    return chanPtr->instanceData;
}

/*
 *----------------------------------------------------------------------
 *
 * FlushChannel --
 *
 *	This function flushes as much of the queued output as is possible
 *	now. If calledFromAsyncFlush is nonzero, it is being called in an
 *	event handler to flush channel output asynchronously.
 *
 * Results:
 *	0 if successful, else the error code that was returned by the
 *	channel type operation.
 *
 * Side effects:
 *	May produce output on a channel. May block indefinitely if the
 *	channel is synchronous. May schedule an async flush on the channel.
 *	May recycle memory for buffers in the output queue.
 *
 *----------------------------------------------------------------------
 */

static int
FlushChannel(chanPtr, calledFromAsyncFlush)
    TclChannel *chanPtr;		/* The channel on which to flush. */
    int calledFromAsyncFlush;		/* If non-zero then we are being
                                         * called from an async flush
                                         * callback. */
{
    int written;			/* How much output was written? */
    int toWrite;			/* How much output to write? */
    int errorCode;			/* Error code from channel level
                                         * output function. */
    ChannelBuffer *curOutPtr;		/* Current output buffer being
                                         * flushed to the channel. */
    ChannelBuffer *savedBuffer;		/* Buffer to replace current
                                         * output buffer being filled; saved
                                         * in the queue-emptying loop to
                                         * eliminate the need for one
                                         * ckfree/ckalloc sequence. */
    int bufSize;			/* Preferred size for buffer to
                                         * allocate. */
    CloseCallback *cbPtr, *cbNextPtr;	/* Iterate over close callbacks
                                         * for this channel. */
    TclChannel *prevChanPtr;		/* Preceding channel in list of
                                         * all channels - used to splice a
                                         * channel out of the list on close. */
    
    savedBuffer = (ChannelBuffer *) NULL;
    errorCode = 0;

    bufSize = (chanPtr->typePtr->bufferSizeProc) (chanPtr->instanceData);
    
    /*
     * If an async flush is active and we are not being called from it,
     * let the async flush take care of any queued input. We only move
     * the current output buffer to the queue, if it is ready or full,
     * and allocate a new buffer.
     */
    
    if ((!calledFromAsyncFlush) && (chanPtr->flags & TCL_ASYNC_FLUSH_ACTIVE)) {
        if ((chanPtr->curOutPtr != (ChannelBuffer *) NULL) &&
                ((chanPtr->flags & TCL_BUFFER_READY) ||
                        (chanPtr->curOutPtr->nextAdded ==
                                chanPtr->curOutPtr->bufSize))) {
            
            chanPtr->flags &= (~(TCL_BUFFER_READY));

            if (chanPtr->outQueueTail == (ChannelBuffer *) NULL) {
                chanPtr->outQueueHead = chanPtr->curOutPtr;
            } else {
                chanPtr->outQueueTail->nextPtr = chanPtr->curOutPtr;
            }
            chanPtr->outQueueTail = chanPtr->curOutPtr;

            chanPtr->curOutPtr = (ChannelBuffer *) ckalloc((unsigned)
                    CHANNELBUFFER_HEADER_SIZE + bufSize);
            chanPtr->curOutPtr->nextAdded = 0;
            chanPtr->curOutPtr->nextRemoved = 0;
            chanPtr->curOutPtr->bufSize = bufSize;
            chanPtr->curOutPtr->nextPtr = (ChannelBuffer *) NULL;
        }
        return 0;
    }
    
    /*
     * Loop trying to write out one buffer at a time until no more ready output
     * remains (either in the queue or in the current output buffer), or until
     * an error occurs.
     */
top:    
    while (1) {
        
        /*
         * Attempt to move the current output buffer into the output queue
         * only when the queue is empty and the buffer is ready.
         */
        
        if ((chanPtr->outQueueHead == (ChannelBuffer *) NULL) &&
                ((chanPtr->curOutPtr != (ChannelBuffer *) NULL) &&
                        ((chanPtr->flags & TCL_BUFFER_READY) ||
                                (chanPtr->curOutPtr->nextAdded ==
                                        chanPtr->curOutPtr->bufSize)))) {

            chanPtr->flags &= (~(TCL_BUFFER_READY));
            chanPtr->curOutPtr->nextPtr = (ChannelBuffer *) NULL;
            chanPtr->outQueueHead = chanPtr->outQueueTail = chanPtr->curOutPtr;
            if (savedBuffer == (ChannelBuffer *) NULL) {
                chanPtr->curOutPtr = (ChannelBuffer *) ckalloc((unsigned)
                        CHANNELBUFFER_HEADER_SIZE + bufSize);
                chanPtr->curOutPtr->bufSize = bufSize;
            } else {
                chanPtr->curOutPtr = savedBuffer;
                savedBuffer = (ChannelBuffer *) NULL;
            }
            chanPtr->curOutPtr->nextAdded = 0;
            chanPtr->curOutPtr->nextRemoved = 0;
            chanPtr->curOutPtr->nextPtr = (ChannelBuffer *) NULL;
        }

        /*
         * If the queue is still empty then we have finished producing all
         * the output. Cancel any async flush scheduled for this channel.
         */
        
        if (chanPtr->outQueueHead == (ChannelBuffer *) NULL) {
            if (chanPtr->flags & TCL_ASYNC_FLUSH_ACTIVE) {
                chanPtr->flags &= (~(TCL_ASYNC_FLUSH_ACTIVE));
                if (chanPtr->outFile != (Tcl_File) NULL) {
                    Tcl_DeleteFileHandler(chanPtr->outFile);
                }
            }
            break;
        }

        /*
         * OK, attempt to write out the contents of the buffer at the head of
         * the output queue.
         */

        curOutPtr = chanPtr->outQueueHead;
        toWrite = curOutPtr->nextAdded - curOutPtr->nextRemoved;

        written = (chanPtr->typePtr->outputProc) (chanPtr->instanceData,
                chanPtr->outFile, curOutPtr->buf + curOutPtr->nextRemoved,
                toWrite, &errorCode);
            
	/*
         * If the write failed completely attempt to start the asynchronous
         * flush mechanism and break out of this loop - do not attempt to
         * write any more output at this time.
         */
        
        if (written < 0) {
            if (errorCode == EINTR) {
                continue;
            }
            if (((errorCode == EWOULDBLOCK) || (errorCode == EAGAIN)) &&
                    (chanPtr->flags & TCL_ASYNC_CHANNEL)) {
            
                if (!(chanPtr->flags & TCL_ASYNC_FLUSH_ACTIVE)) {
                    Tcl_CreateFileHandler(chanPtr->outFile, TCL_WRITABLE,
                        ChannelEventDispatcher, (ClientData) chanPtr);
                }
                chanPtr->flags |= TCL_ASYNC_FLUSH_ACTIVE;
                errorCode = 0;
                break;

            }

            /*
             * We do not know how much output was actually written, so
             * we assume that all of it was written.
             */
                
            written = toWrite;

            /*
             * Decide whether to report the error upwards or defer it.
             */

            if (calledFromAsyncFlush) {
                if (chanPtr->latentError == 0) {
                    chanPtr->latentError = errorCode;
                }
            } else {
                Tcl_SetErrno(errorCode);
            }
        }
        
        curOutPtr->nextRemoved += written;

        /*
         * If we have produced all output queued in this buffer then
         * remove this buffer from the queue.
         */
        
        if (curOutPtr->nextRemoved == curOutPtr->nextAdded) {
            chanPtr->outQueueHead = chanPtr->outQueueHead->nextPtr;
            if (chanPtr->outQueueHead == (ChannelBuffer *) NULL) {
                chanPtr->outQueueTail = (ChannelBuffer *) NULL;
            }

            /*
             * Save one buffer around to replace chanPtr->curOutPtr
             * if we decide to move it into the queue.
             */
                
            if (savedBuffer == (ChannelBuffer *) NULL) {
                savedBuffer = curOutPtr;
            } else {
                ckfree((char *) curOutPtr);
            }
        }
    }		/* Closes "while" */
    
    /*
     * Now check if the current buffer (if any) is full, or if the channel
     * is closed and the current buffer (if any) has any output,
     * and if so, move it into the queue and install a new one.
     */

    if ((chanPtr->curOutPtr != (ChannelBuffer *) NULL) &&
            ((chanPtr->curOutPtr->nextAdded == chanPtr->curOutPtr->bufSize) ||
                ((chanPtr->flags & TCL_CHANNEL_CLOSED) &&
                        (chanPtr->curOutPtr->nextAdded != 0)))) {

        chanPtr->flags &= (~(TCL_BUFFER_READY));
        chanPtr->curOutPtr->nextPtr = chanPtr->outQueueTail;
        chanPtr->outQueueTail = chanPtr->curOutPtr;
        if (chanPtr->outQueueHead == (ChannelBuffer *) NULL) {
            chanPtr->outQueueHead = chanPtr->outQueueTail;
        }
        if (savedBuffer == (ChannelBuffer *) NULL) {
            chanPtr->curOutPtr = (ChannelBuffer *) ckalloc((unsigned)
                    CHANNELBUFFER_HEADER_SIZE + bufSize);
            chanPtr->curOutPtr->bufSize = bufSize;
        } else {
            chanPtr->curOutPtr = savedBuffer;
            savedBuffer = (ChannelBuffer *) NULL;
        }
        chanPtr->curOutPtr->nextAdded = 0;
        chanPtr->curOutPtr->nextRemoved = 0;
        chanPtr->curOutPtr->nextPtr = (ChannelBuffer *) NULL;
        goto top;
    }
    
    /*
     * If a saved buffer remains that was not reused, release its storage now.
     */

    if (savedBuffer != (ChannelBuffer *) NULL) {
        ckfree((char *) savedBuffer);
    }
    
    /*
     * If the channel is closed, delete it when the refcount drops to zero.
     */
    
    if (chanPtr->flags & TCL_CHANNEL_CLOSED) {

        /*
         * If the TCL_CHANNEL_CLOSED flag is set, no more input can be
         * retrieved from the channel, so it is safe to discard it now.
         */

        DiscardInputQueued(chanPtr);

        /*
         * If there is an async flush mechanism active then let it finish the
         * flush before destroying the actual channel.
         */

        if (chanPtr->flags & TCL_ASYNC_FLUSH_ACTIVE) {            
            return errorCode;
        }

        /*
         * If there is an empty buffer in curOutPtr, free its storage.
         */

        if ((chanPtr->curOutPtr != (ChannelBuffer *) NULL) &&
                (chanPtr->curOutPtr->nextAdded == 0)) {
            ckfree((char *) chanPtr->curOutPtr);
            chanPtr->curOutPtr = (ChannelBuffer *) NULL;
        }
    
        /*
         * If we get here then it must be true that there are no more buffers
         * queued for output and there is no buffer in curOutPtr.
         */

        if ((chanPtr->outQueueHead != (ChannelBuffer *) NULL) ||
                (chanPtr->curOutPtr != (ChannelBuffer *) NULL) ||
                (chanPtr->inQueueHead != (ChannelBuffer *) NULL)) {
            panic("TclFlush, closed channel: queued input or output left");
        }

        /*
         * Remove TCL_READABLE and TCL_WRITABLE from chanPtr->flags, so
         * that close callbacks can not do input or output (assuming they
         * squirreled the channel away in their clientData). This also
         * prevents infinite loops if the callback calls any C API that
         * could call FlushChannel.
         */

        chanPtr->flags &= (~(TCL_READABLE|TCL_WRITABLE));
        
        /*
         * Invoke the registered close callbacks and delete their records.
         * It is the job of each callback to dispose of its own callbackData.
         */

        for (cbPtr = chanPtr->closeCallbacks;
                 cbPtr != (CloseCallback *) NULL;
                 cbPtr = cbNextPtr) {
            cbNextPtr = cbPtr->nextPtr;
            (cbPtr->proc) (cbPtr->clientData);
            ckfree((char *) cbPtr);
        }
        
        /*
         * Splice this channel out of the list of all channels.
         */

        if (chanPtr == firstChanPtr) {
            firstChanPtr = chanPtr->nextChanPtr;
        } else {
            for (prevChanPtr = firstChanPtr;
                     (prevChanPtr != (TclChannel *) NULL) &&
                         (prevChanPtr->nextChanPtr != chanPtr);
                     prevChanPtr = prevChanPtr->nextChanPtr) {
                /* Empty loop body. */
            }
            if (prevChanPtr == (TclChannel *) NULL) {
                panic("FlushChannel: damaged channel list");
            }
            prevChanPtr->nextChanPtr = chanPtr->nextChanPtr;
        }
        
        /*
         * OK, close the channel itself.
         */
        
        errorCode = (chanPtr->typePtr->closeProc) (chanPtr->instanceData,
                chanPtr->inFile, chanPtr->outFile);
    
        /*
         * If we are being called synchronously, report either
         * any latent error on the channel or the current error.
         */
        
        if (!calledFromAsyncFlush) {
            if (chanPtr->latentError != 0) {
                errorCode = chanPtr->latentError;
            }
            if (errorCode != 0) {
                Tcl_SetErrno(errorCode);
            }
        }

        if (chanPtr->channelName != (char *) NULL) {
            ckfree(chanPtr->channelName);
        }

        ckfree((char *) chanPtr);
    }

    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_Write --
 *
 *	Puts a sequence of characters into an output buffer, may queue the
 *	buffer for output if it gets full, and also remembers whether the
 *	current buffer is ready (it contains a newline and we are in line
 *	buffering mode). This is a public function so that other C
 *	facilities can be built on top of it.
 *
 * Results:
 *	The number of bytes written or -1. If -1, Tcl_GetErrno will
 *	return the error code.
 *
 * Side effects:
 *	May buffer up output and may cause output to be produced on a
 *	specific channel.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_Write(chan, src, slen)
    Tcl_Channel chan;			/* The channel to buffer output for. */
    char *src;				/* Output to buffer. */
    int slen;				/* Its length. Negative means
                                         * the output is null terminated
                                         * and we have to compute the length
                                         * of the output. */
{
    TclChannel *chanPtr;		/* The internal channel. */
    ChannelBuffer *outBufPtr;		/* Current output buffer. */
    int foundNewline;			/* Did we find a newline in output? */
    char *d, *s;			/* Search variables for newline. */
    int crsent;				/* In network eol translation mode,
                                         * remember the fact that a CR was
                                         * output to the channel without
                                         * its following NL. */
    int i;				/* Loop index for newline search. */
    int toCopy;				/* How many characters will fit in
                                         * the current buffer (after eol
                                         * translation)? */
    int copied;				/* How many characters were actually
                                         * copied from the source string
                                         * into the current buffer? */
    int totalCopied;			/* How many characters total were
                                         * copied to the channel buffer? */
    char *dest;				/* Where in line to copy to? */
    int bufSize;			/* Preferred size for buffers. */

    chanPtr = (TclChannel *) chan;

    /*
     * Check for latent unreported error.
     */

    if (chanPtr->latentError != 0) {
        Tcl_SetErrno(chanPtr->latentError);
        chanPtr->latentError = 0;
        return -1;
    }
    
    /*
     * If the channel is not open for writing punt.
     */

    if (!(chanPtr->flags & TCL_WRITABLE)) {
        Tcl_SetErrno(EACCES);
        return -1;
    }
    
    /*
     * If length passed is negative, assume that the output is null terminated
     * and compute its length.
     */
    
    if (slen < 0) {
        slen = strlen(src);
    }
    
    /*
     * If we are in network (or windows) translation mode, record the fact
     * that we have not yet sent a CR to the channel.
     */

    crsent = 0;
    
    /*
     * If this is the first write and the channel is still in unknown
     * end-of-line translation mode, force the mode to be the platform
     * specific mode for each platform.
     */

    if (chanPtr->translation == TCL_TRANSLATE_AUTO) {
        chanPtr->translation = TCL_PLATFORM_TRANSLATION;
    }
    
    /*
     * Loop filling buffers and flushing them until all output has been
     * consumed.
     */

    totalCopied = 0;
    bufSize = (chanPtr->typePtr->bufferSizeProc)(chanPtr->instanceData);

    while (slen > 0) {
        
        /*
         * Make sure there is a current output buffer to accept output.
         */

        if (chanPtr->curOutPtr == (ChannelBuffer *) NULL) {
            chanPtr->curOutPtr = (ChannelBuffer *) ckalloc((unsigned)
                    CHANNELBUFFER_HEADER_SIZE + bufSize);
            chanPtr->curOutPtr->nextAdded = 0;
            chanPtr->curOutPtr->nextRemoved = 0;
            chanPtr->curOutPtr->nextPtr = (ChannelBuffer *) NULL;
            chanPtr->curOutPtr->bufSize = bufSize;
        }

        outBufPtr = chanPtr->curOutPtr;

        toCopy = outBufPtr->bufSize - outBufPtr->nextAdded;
        if (toCopy > slen) {
            toCopy = slen;
        }
        
        dest = outBufPtr->buf + outBufPtr->nextAdded;
        switch (chanPtr->translation) {
            case TCL_TRANSLATE_LF:
                copied = toCopy;
                memcpy((VOID *) dest, (VOID *) src, (size_t) toCopy);
                break;
            case TCL_TRANSLATE_CR:
                copied = toCopy;
                memcpy((VOID *) dest, (VOID *) src, (size_t) toCopy);
                for (d = dest; d < dest + toCopy; d++) {
                    if (*d == '\n') {
                        *d = '\r';
                    }
                }
                break;
            case TCL_TRANSLATE_CRLF:
                for (copied = 0, d = dest, s = src;
                     d < dest + toCopy;
                     d++, s++, copied++) {
                    if ((*s == '\n') && (crsent == 0)) {
                        *d = '\r';
                        s--, copied--;
                        crsent = 1;
                    } else if ((*s == '\n') && (crsent == 1)) {
                        *d = '\n';
                        crsent = 0;
                    } else {
                        *d = *s;
                    }
                }
                break;
            case TCL_TRANSLATE_AUTO:
                panic("AUTO mode in output - should not happen");
                /* NOTREACHED */
                return -1;
            default:
                panic("unknown channel eol translation mode");
                /* NOTREACHED */
                return -1;
        }

        /*
         * The current buffer is ready for output if it is full, or if it
         * contains a newline and this channel is line-buffered.
         */

        outBufPtr->nextAdded += toCopy;
        if (!(chanPtr->flags & TCL_BUFFER_READY)) {
            if (outBufPtr->nextAdded == outBufPtr->bufSize) {
                chanPtr->flags |= TCL_BUFFER_READY;
            } else if (chanPtr->flags & TCL_LINE_BUFFERED) {
                for (s = src, i = 0, foundNewline = 0;
                         (i < copied) && (!foundNewline);
                         i++, s++) {

                    /*
                     * Internal canonical representation always uses "\n"
                     * for newline character.
                     */
                    
                    if (*s == '\n') {
                        foundNewline = 1;
                        break;
                    }
                }
                if (foundNewline) {
                    chanPtr->flags |= TCL_BUFFER_READY;
                }
            }
        }
        
        totalCopied += copied;
        src += copied;
        slen -= copied;

        if (chanPtr->flags & TCL_BUFFER_READY) {

            /*
             * The TCL_BUFFER_READY flag is reset inside the
             * FlushChannel call. FlushChannel also calls Tcl_SetErrno()
             * to propagate the error code.
             */
            
            if (FlushChannel(chanPtr, 0) != 0) {
                return -1;
            }
        }
    } /* Closes "while" */

    return totalCopied;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_Flush --
 *
 *	Implement flushing on Tcl channels. This is a public function
 *	so that other C facilities may be built on top of it.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May flush output queued on this channel. May discard input
 *	queued for the channel. May delete the channel if it is closed
 *	and it is the last reference to the channel.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_Flush(chan)
    Tcl_Channel chan;			/* The TclChannel to flush. */
{
    int result;				/* Of calling FlushChannel. */
    TclChannel *chanPtr;		/* The real IO channel. */

    chanPtr = (TclChannel *) chan;

    /*
     * Check for latent unreported error.
     */

    if (chanPtr->latentError != 0) {
        Tcl_SetErrno(chanPtr->latentError);
        chanPtr->latentError = 0;
        return TCL_ERROR;
    }

    /*
     * If the channel is not open for writing punt.
     */

    if (!(chanPtr->flags & TCL_WRITABLE)) {
        Tcl_SetErrno(EACCES);
        return TCL_ERROR;
    }
    
    /*
     * Force current output buffer to be output also.
     */
    
    if ((chanPtr->curOutPtr != (ChannelBuffer *) NULL) &&
            (chanPtr->curOutPtr->nextAdded > 0)) {
        chanPtr->flags |= TCL_BUFFER_READY;
    }
    
    result = FlushChannel(chanPtr, 0);
    if (result != 0) {
        return TCL_ERROR;
    }

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DiscardInputQueued --
 *
 *	Discards any input read from the channel but not yet consumed
 *	by Tcl reading commands.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	May discard input from the channel.
 *
 *----------------------------------------------------------------------
 */

static void
DiscardInputQueued(chanPtr)
    TclChannel *chanPtr;		/* Channel on which to discard
                                 * the queued input. */
{
    ChannelBuffer *runPtr, *nxtPtr;	/* Loop variables. */

    for (runPtr = chanPtr->inQueueHead;
             runPtr != (ChannelBuffer *) NULL;
             runPtr = nxtPtr) {
        nxtPtr = runPtr->nextPtr;
        ckfree((char *) runPtr);
    }
    chanPtr->inQueueHead = (ChannelBuffer *) NULL;
    chanPtr->inQueueTail = (ChannelBuffer *) NULL;    
}

/*
 *----------------------------------------------------------------------
 *
 * GetInputBuffer --
 *
 *	Gets a new input buffer from the underlying channel.
 *
 * Results:
 *	A Posix error code or 0.
 *
 * Side effects:
 *	Reads from the underlying channel. May update the current
 *	access pointer.
 *
 *----------------------------------------------------------------------
 */

static int
GetInputBuffer(chanPtr, toRead)
    TclChannel *chanPtr;		/* Channel to read input from. */
    int toRead;				/* If > 0, specifies how much
                	                 * input to read (upto the size of
                        	         * one buffer). If <= 0, read as much
                                	 * as will fit in a buffer. */
{
    int result;				/* Result of channel type function. */
    int nread;				/* How much was read from channel? */
    ChannelBuffer *bufPtr;		/* New buffer to add to input queue. */
    int bufSize;			/* Preferred size of buffers. */
    
    /*
     * If the last read left us in a pending-eof state produce a buffer
     * containing just the "\r" and return.
     */

    if ((chanPtr->flags & TCL_INPUT_EOF_PENDING) &&
            (chanPtr->translation == TCL_TRANSLATE_CRLF)) {
        chanPtr->flags &= (~(TCL_INPUT_EOF_PENDING));
        chanPtr->flags |= TCL_CHANNEL_EOF;
        return 0;
    }
    
    /*
     * Compute how much to read: if asked to read -1 bytes, read as much
     * as will fit in a buffer. If asked to read a positive number of
     * bytes, read the minimum of that request and the size of a buffer.
     */

    bufSize = (chanPtr->typePtr->bufferSizeProc) (chanPtr->instanceData);
    if (toRead <= 0) {
        toRead = bufSize;
    } else if (toRead > bufSize) {
        toRead = bufSize;
    }

    bufPtr = (ChannelBuffer *) ckalloc(
        (unsigned) CHANNELBUFFER_HEADER_SIZE + bufSize);
    bufPtr->nextRemoved = 0;
    bufPtr->nextAdded = 0;
    bufPtr->bufSize = bufSize;
    
    while (1) {
        nread = (chanPtr->typePtr->inputProc) (chanPtr->instanceData,
                chanPtr->inFile, bufPtr->buf, toRead, &result);
        if (result != EINTR) {
            break;
        }
    }

    if (nread < 0) {
        ckfree((char *) bufPtr);
        if (result == EWOULDBLOCK) {
            result = EAGAIN;
        }
        Tcl_SetErrno(result);
        return result;
    }

    /*
     * If zero read, assume (transient?) EOF. This is signalled up as
     * setting the flags TCL_CHANNEL_EOF and returning a read of 0 bytes.
     */
    
    if (nread == 0) {

        /*
         * If the CR seen indicator is set and we are in CRLF translation
         * mode then produce a "\r" and behave as though this was just
         * read from the channel. This ensures that a "\r" immediately
         * followed by EOF will still be seen by Tcl code.
         */

        if ((chanPtr->flags & TCL_INPUT_SAW_CR) &&
                (chanPtr->translation == TCL_TRANSLATE_CRLF)) {

            /*
             * DO NOT set TCL_CHANNEL_EOF here - the callers depend on it
             * not being set if a buffer was produced. Instead set a special
             * state bit (TCL_INPUT_EOF_PENDING) which is seen before the
             * next read and converted into TCL_CHANNEL_EOF then.
             */

            chanPtr->flags |= TCL_INPUT_EOF_PENDING;
            bufPtr->buf[0] = '\r';
            nread = 1;
        } else {
            ckfree((char *) bufPtr);
            chanPtr->flags |= TCL_CHANNEL_EOF;
            return 0;
        }
    }

    bufPtr->nextAdded = nread;
    
    if (chanPtr->inQueueTail == (ChannelBuffer *) NULL) {
        chanPtr->inQueueHead = bufPtr;
    } else {
        chanPtr->inQueueTail->nextPtr = bufPtr;
    }
    chanPtr->inQueueTail = bufPtr;
    bufPtr->nextPtr = (ChannelBuffer *) NULL;

    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * CopyAndTranslateInput --
 *
 *	Copy input from buffer queue into destination string, and apply
 *	eol translations according to mode in effect currently.
 *
 * Results:
 *	Number of characters (as opposed to bytes) copied.
 *
 * Side effects:
 *	Consumes buffered input.
 *
 *----------------------------------------------------------------------
 */

static int
CopyAndTranslateInput(chanPtr, result, space)
    TclChannel *chanPtr;	/* The channel from which to consume
                                 * the input. */
    char *result;		/* Where to store the copied input. */
    int space;			/* How many bytes are available in result
                                 * to store the copied input? */
{
    int bufSize;		/* How many bytes are available to be
                                 * copied in the current input buffer? */
    size_t toCopyNow;		/* How many bytes should be copied from
                                 * the current buffer? This may be
                                 * <= bytesInThisBuffer. */
    ChannelBuffer *bufPtr;
    char *r;			/* Always points to the next free location
				 * in the result buffer. */
    char *s;			/* Loop variable for crlf mode. */
    char *prevEndPtr;		/* Result was filled up to here by
                                 * copying from the previous buffer. */
    char curByte;		/* The byte we are currently translating. */
    
    r = result;

    /*
     * Translate input from channel buffers until there are no more buffers
     * or the result buffer is filled.
     */

    while ((chanPtr->inQueueHead != NULL) && (space > 0)) {

        /*
         * Store up to where we filled result in prevEndPtr.
         */

        prevEndPtr = r;
        
	/*
	 * Take the first buffer on the queue and compute the amount of
	 * unread data in the buffer that we should copy.
	 */

	bufPtr = chanPtr->inQueueHead;
	bufSize = bufPtr->nextAdded - bufPtr->nextRemoved;
	toCopyNow = (bufSize > space) ? space : bufSize;

	/*
	 * Tranlate the next chunk of characters according to the current
	 * translation mode.
	 */

	switch (chanPtr->translation) {
	    case TCL_TRANSLATE_LF:

		/*
		 * Copy the current chunk into the result buffer.
		 */

		memcpy((VOID *) r, (VOID *)(bufPtr->buf + bufPtr->nextRemoved),
                        (size_t) toCopyNow);
		bufPtr->nextRemoved += toCopyNow;
		r += toCopyNow;
		break;

	    case TCL_TRANSLATE_CR:

		/*
		 * Copy the current chunk into the result buffer, then
		 * replace all \r with \n.
		 */

		memcpy((VOID *) r, (VOID *)(bufPtr->buf + bufPtr->nextRemoved),
                        (size_t) toCopyNow);
		bufPtr->nextRemoved += toCopyNow;
		for (; toCopyNow > 0; r++, toCopyNow--) {
		    if (*r == '\r') {
			*r = '\n';
		    }
		}
		break;

	    case TCL_TRANSLATE_CRLF: {
		char *endOfBuf;

		/*
		 * Copy the current chunk into the result buffer, replacing
		 * \r\n with \n.  Note that since there may be a queued CR from
		 * a previous pass, we need to watch for the destination
		 * filling up before we have copied all of the available data.
		 * So we keep track of the amount copied by decrementing
		 * toCopyNow, and we keep track of our position in the buffer
		 * by remembering the last character we are allowed to copy.
		 */

		s = bufPtr->buf + bufPtr->nextRemoved;
		endOfBuf = bufPtr->buf + bufPtr->nextAdded;
		toCopyNow = space;

		while ((s < endOfBuf) && (toCopyNow > 0)) {
		    if (*s == '\r') {

			/*
			 * If the previous character was a CR, then insert
			 * a \r into the result, otherwise wait until we see
			 * what the next character is before inserting the \r.
			 */

			if (chanPtr->flags & TCL_INPUT_SAW_CR) {
			    *r++ = '\r';
			    toCopyNow--;
			} else {
			    chanPtr->flags |= TCL_INPUT_SAW_CR;
			}
			s++;
		    } else {

			/*
			 * If the current character is a NL, then we just
			 * insert it and advance the source pointer.
			 * Otherwise, if the previous character was a CR, then
			 * we insert it and leave the source pointer on the
			 * current character so we can pick it up on the next
			 * iteration.
			 */

			if (*s == '\n') {
			    *r++ = '\n';
			    s++;
			} else if (chanPtr->flags & TCL_INPUT_SAW_CR) {
			    *r++ = '\r';
			} else {
			    *r++ = *s++;
			}
			toCopyNow--;
			chanPtr->flags &= ~(TCL_INPUT_SAW_CR);
		    }
		}
		bufPtr->nextRemoved = s - bufPtr->buf;
		break;
            }

	    case TCL_TRANSLATE_AUTO:
		s = bufPtr->buf + bufPtr->nextRemoved;
		
		/*
		 * Look for the first end of line character.  Once we find it,
		 * stop processing the current chunk and restart from the
		 * top so we can process the rest in the appropriate platform
		 * specific manner.
		 */

                for (; toCopyNow > 0; ) {
                    toCopyNow--;
                    curByte = *s++;
		    if (curByte == '\n') {
			if (chanPtr->flags & TCL_INPUT_SAW_CR) {
                            chanPtr->flags &= ~(TCL_INPUT_SAW_CR);
			    chanPtr->translation = TCL_TRANSLATE_CRLF;
			} else {
                            *r++ = curByte;
			    chanPtr->translation = TCL_TRANSLATE_LF;
			}
                        break;				/* FOR */
		    } else if (curByte == '\r') {
                        *r++ = '\n';

                        /*
                         * If the previous character was a CR we can
                         * switch to CR mode.
                         */
                        
			if (chanPtr->flags & TCL_INPUT_SAW_CR) {
                            chanPtr->flags &= (~(TCL_INPUT_SAW_CR));
			    chanPtr->translation = TCL_TRANSLATE_CR;
			    break;			/* FOR */
			} else {
			    chanPtr->flags |= TCL_INPUT_SAW_CR;
			}
		    } else {
                        *r++ = curByte;
			if (chanPtr->flags & TCL_INPUT_SAW_CR) {
			    chanPtr->translation = TCL_TRANSLATE_CR;
			    chanPtr->flags &= ~(TCL_INPUT_SAW_CR);
                            break;			/* FOR */
			}
		    }
		}
		bufPtr->nextRemoved = s - bufPtr->buf;
		break;

	    default:
		panic("unknown eol translation mode");
	}

	/*
	 * Adjust the amount of space left based on the current position
	 * of the r pointer.
	 */

	space -= (r - prevEndPtr);

	/*
	 * Discard the current buffer if it is empty.
	 */

	if (bufPtr->nextRemoved == bufPtr->nextAdded) {
	    chanPtr->inQueueHead = bufPtr->nextPtr;
	    if (chanPtr->inQueueHead == NULL) {
		chanPtr->inQueueTail = NULL;
	    }
	    ckfree((char*) bufPtr);
	}
    }

    /*
     * Return the number of characters copied into the result buffer.
     */

    return (r - result);
}

/*
 *----------------------------------------------------------------------
 *
 * GetNewline --
 *
 *	Accumulate input into the channel input buffer queue until a
 *	newline has been seen.
 *
 * Results:
 *	Number of bytes buffered or -1 on failure.
 *
 * Side effects:
 *	Consumes input from the channel.
 *
 *----------------------------------------------------------------------
 */

static int
GetNewline(chanPtr)
    TclChannel *chanPtr;		/* Channel to queue input on. */
{
    int bytesQueued;		/* How much input is already queued? */
    int result;			/* Of getting another buffer from the
                                 * channel. */
    ChannelBuffer *run, *prev;	/* Iterate over queued buffers. */
    char *r;			/* Iterates over input in one buffer. */
    int newlineFound;		/* Have we found a newline yet? */
    int bytesToNewline;		/* How many bytes in buffer up to and
                                 * including the newline? */
    int crseen;			/* If in network translation mode, did we
                                 * see a "\r" as the immediately preceding
                                 * character? */

    /*
     * Clear EOF and BLOCKED bits because we want to discover this anew
     * during each operation.
     */

    chanPtr->flags &= (~(TCL_CHANNEL_EOF | TCL_CHANNEL_BLOCKED));
    
    bytesToNewline = 0;
    crseen = 0;

    for (run = chanPtr->inQueueHead, prev = (ChannelBuffer *) NULL,
             newlineFound = 0, bytesQueued = 0;
             !newlineFound;
             run = prev->nextPtr, bytesQueued += bytesToNewline) {

        /*
         * If we reached the end of the queue, try to get one more buffer
         * of input.
         */
        
        if (run == (ChannelBuffer *) NULL) {
            result = GetInputBuffer(chanPtr, -1);
            if (result != 0) {
                if (result == EAGAIN) {
                    chanPtr->flags |= TCL_CHANNEL_BLOCKED;

                    /*
                     * Signal the blocking condition as a -1 return value.
                     * NOTE that this is different than in Tcl_Read where
                     * a blocking condition is signalled as a return value
                     * of 0.
                     */
                    
                    return -1;
                }
                DiscardInputQueued(chanPtr);
                return -1;
            }
            if (chanPtr->flags & TCL_CHANNEL_EOF) {

                /*
                 * If we discovered EOF when attempting to read the first
                 * byte, signal this as -1. Otherwise allow Tcl_Gets to
                 * see the partial line; the next call to GetNewLine will
                 * return -1 if the EOF persists.
                 *
                 * NOTE that this is different than in Tcl_Read where
                 * an EOF condition is signalled as a return value of 0.
                 */
                
                if (bytesQueued == 0) {
                    return -1;
                }
                return bytesQueued;
            }
            if (prev == (ChannelBuffer *) NULL) {
                run = chanPtr->inQueueHead;
            } else {
                run = prev->nextPtr;
            }
        }

        prev = run;
        for (r = run->buf + run->nextRemoved, bytesToNewline = 0,
                 newlineFound = 0;
                 (r < run->buf + run->nextAdded) && (!newlineFound);
                 r++, bytesToNewline++) {
            switch (chanPtr->translation) {
                case TCL_TRANSLATE_AUTO:
                    if (*r == '\n') {
                        if (chanPtr->flags & TCL_INPUT_SAW_CR) {
			    bytesToNewline--;
			} else {
			    newlineFound = 1;
			}
                    } else if (*r == '\r') {
                        newlineFound = 1;
                    }
                    break;
                case TCL_TRANSLATE_LF:
                    if (*r == '\n') {
                        newlineFound = 1;
                    }
                    break;
                case TCL_TRANSLATE_CR:
                    if (*r == '\r') {
                        newlineFound = 1;
                    }
                    break;
                case TCL_TRANSLATE_CRLF:
                    if ((*r == '\n') && (crseen)) {
                        newlineFound = 1;

			/*
			 * CopyAndTranslate wants to know the length of the
			 * result, not the input.  The input is one larger
			 * because crlf shrinks to lf.
			 */

			bytesToNewline--;
                    } else if (*r == '\r') {
                        crseen = 1;
                    } else {
                        crseen = 0;
                    }
                    break;
                default:
                    panic("unknown eol translation mode");
            }
        }
    }

    return bytesQueued;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_Read --
 *
 *	Implements the work of reading a given number of characters from
 *	an IO channel. This function is public so that other C facilities
 *	can be built on top of it.
 *
 * Results:
 *	The number of characters read, or -1 on error. Use Tcl_GetErrno()
 *	to retrieve the error code for the error that occurred.
 *
 * Side effects:
 *	May cause output to be flushed. May consume input from channel.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_Read(chan, bufPtr, toRead)
    Tcl_Channel chan;		/* The channel from which to read. */
    char *bufPtr;		/* Where to store input read. Storage
                                 * is allocated in this function and must
                                 * be released by caller. */
    int toRead;			/* Maximum number of characters to read. */
{
    TclChannel *chanPtr;	/* The real IO channel. */
    int copied;			/* How many characters were copied into
                                 * the result string? */
    int result;			/* Of calling GetInputBuffer. */
    
    chanPtr = (TclChannel *) chan;

    /*
     * Punt if the channel is not opened for reading.
     */

    if (!(chanPtr->flags & TCL_READABLE)) {
        Tcl_SetErrno(EACCES);
        return -1;
    }
    
    /*
     * Clear EOF and BLOCKED bits because we want to discover this anew
     * during each operation.
     */

    chanPtr->flags &= (~(TCL_CHANNEL_EOF | TCL_CHANNEL_BLOCKED));

    for (copied = CopyAndTranslateInput(chanPtr, bufPtr, toRead);
             copied < toRead;
             copied += CopyAndTranslateInput(chanPtr, bufPtr + copied,
                     toRead - copied)) {
        result = GetInputBuffer(chanPtr, toRead - copied);
        if (result != 0) {
            if (result == EWOULDBLOCK) {
                chanPtr->flags |= TCL_CHANNEL_BLOCKED;
                return copied;
            }
            DiscardInputQueued(chanPtr);
            return -1;
        }
        if (chanPtr->flags & TCL_CHANNEL_EOF) {
            break;
        }
    }
    return copied;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_Gets --
 *
 *	Implements reading from a channel until the first newline. The
 *	number of characters or -1 is returned in an output variable, and the
 *	string read is returned through the dynamic string structure passed
 *	in as argument.
 *
 * Results:
 *	Length of line read or -1 if error. If -1, use Tcl_GetErrno() to
 *	retrieve the POSIX error code for the error that occurred.
 *
 * Side effects:
 *	May flush output on the channel. May cause input to be
 *	consumed from the channel.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_Gets(chan, lineRead)
    Tcl_Channel chan;		/* Channel from which to read. */
    Tcl_DString *lineRead;	/* Dynamic string for storing the
                                 * line to be read. */
{
    TclChannel *chanPtr;	/* The channel to read from. */
    char *buf;			/* Points into DString where data
                                 * will be stored. */
    int offset;			/* Offset from start of DString at
                                 * which to append the line just read. */
    int lineLen;		/* Length of line just read. */
    int bytesBuffered;		/* How many bytes are there in the
                                 * current line? */

    chanPtr = (TclChannel *) chan;

    /*
     * Punt if the channel is not opened for reading.
     */

    if (!(chanPtr->flags & TCL_READABLE)) {
        Tcl_SetErrno(EACCES);
        return -1;
    }
    
    /*
     * Clear EOF and BLOCKED bits because we want to discover this anew
     * during each operation.
     */

    chanPtr->flags &= (~(TCL_CHANNEL_EOF | TCL_CHANNEL_BLOCKED));

    bytesBuffered = GetNewline(chanPtr);
    if (bytesBuffered == -1) {
        return -1;
    }
    if (bytesBuffered == 0) {
        return 0;
    }

    offset = Tcl_DStringLength(lineRead);
    Tcl_DStringSetLength(lineRead, bytesBuffered + offset);
    buf = Tcl_DStringValue(lineRead) + offset;
    lineLen = CopyAndTranslateInput(chanPtr, buf, bytesBuffered);
    if ((lineLen > 0) && (buf[lineLen - 1] == '\n')) {
        lineLen--;
    }
    Tcl_DStringSetLength(lineRead, lineLen + offset);
    return lineLen;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_Seek --
 *
 *	Implements seeking on Tcl TclChannels. This is a public function
 *	so that other C facilities may be implemented on top of it.
 *
 * Results:
 *	The new access point or -1 on error. If error, use Tcl_GetErrno()
 *	to retrieve the POSIX error code for the error that occurred.
 *
 * Side effects:
 *	May flush output on the channel. May discard queued input.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_Seek(chan, offset, mode)
    Tcl_Channel chan;		/* The channel on which to seek. */
    int offset;			/* Offset to seek to. */
    int mode;			/* Relative to which location to seek? */
{
    TclChannel *chanPtr;	/* The real IO channel. */
    ChannelBuffer *run;		/* Iterates over queued buffers. */
    int result;			/* Result of channel type operations. */
    int curPos;			/* Current logical access position in
                                 * the channel. */
    int newPos;			/* New logical position in the channel. */
    int wasAsync;		/* Was this channel asynchronous before
                                 * this function call? If so, we must put
                                 * it back into async output mode when
                                 * we are done. */
    int inputBuffered,
        outputBuffered;		/* How much input and output was buffered? */

    chanPtr = (TclChannel *) chan;

    /*
     * Check for latent unreported error.
     */

    if (chanPtr->latentError != 0) {
        Tcl_SetErrno(chanPtr->latentError);
        chanPtr->latentError = 0;
        return -1;
    }

    /*
     * Disallow seek on channels that are open for neither writing nor
     * reading (e.g. socket server channels).
     */

    if (!(chanPtr->flags & (TCL_WRITABLE|TCL_READABLE))) {
        Tcl_SetErrno(EACCES);
        return -1;
    }

    /*
     * Disallow seek on channels whose type does not have a seek procedure
     * defined. This means that the channel does not support seeking.
     */

    if (chanPtr->typePtr->seekProc == (Tcl_DriverSeekProc *) NULL) {
        Tcl_SetErrno(EINVAL);
        return -1;
    }

    /*
     * Compute how much input or output is buffered, to enable subsequent
     * computation of the current logical access position.
     */

    inputBuffered = 0;
    outputBuffered = 0;
    if (mode == SEEK_CUR) {
        outputBuffered = 0;
        if (chanPtr->curOutPtr != (ChannelBuffer *) NULL) {
            outputBuffered = chanPtr->curOutPtr->nextAdded -
                chanPtr->curOutPtr->nextRemoved;
        }
        for (run = chanPtr->outQueueHead;
                run != (ChannelBuffer *) NULL;
                run = run->nextPtr) {
            outputBuffered += (run->nextAdded - run->nextRemoved);
        }

        inputBuffered = 0;
        for (run = chanPtr->inQueueHead;
                run != (ChannelBuffer *) NULL;
                run = run->nextPtr) {
            inputBuffered += (run->nextAdded - run->nextRemoved);
        }
    }
    
    /*
     * If the channel is in asynchronous output mode, switch it back
     * to synchronous mode and cancel any async flush that may be
     * scheduled. After the flush, the channel will be put back into
     * asynchronous output mode.
     */

    wasAsync = 0;
    if (chanPtr->flags & TCL_ASYNC_CHANNEL) {
        wasAsync = 1;
        result = (chanPtr->typePtr->blockModeProc) (chanPtr->instanceData,
                chanPtr->inFile, chanPtr->outFile, TCL_MODE_BLOCKING);
        if (result != 0) {
            Tcl_SetErrno(result);
            return -1;
        }
        chanPtr->flags &= (~(TCL_ASYNC_CHANNEL));
        if (chanPtr->flags & TCL_ASYNC_FLUSH_ACTIVE) {
            Tcl_DeleteFileHandler(chanPtr->outFile);
            chanPtr->flags &= (~(TCL_ASYNC_FLUSH_ACTIVE));
        }
    }

    /*
     * Record the channel logical access position so that we can seek back
     * to it before doing the actual seek. The logical access position is
     * computed by taking the physical access position and subtracting the
     * amount of input buffered, OR by taking the current physical access
     * position and adding the amount of output still buffered. It is an
     * error if both input and output are buffered.
     */

    curPos = (chanPtr->typePtr->seekProc) (chanPtr->instanceData,
            chanPtr->inFile, chanPtr->outFile, (long) 0, SEEK_CUR, &result);
    if (curPos == -1) {
        Tcl_SetErrno(result);
    } else {
        if ((inputBuffered != 0) && (outputBuffered != 0)) {
            Tcl_SetErrno(EFAULT);
            curPos = -1;
        } else if (inputBuffered != 0) {
            curPos -= inputBuffered;
        } else if (outputBuffered != 0) {
            curPos += outputBuffered;
        }
    }

    /*
     * If there is a partially filled output buffer, flush that also.
     */

    if ((chanPtr->curOutPtr != (ChannelBuffer *) NULL) &&
            (chanPtr->curOutPtr->nextAdded > 0)) {
        chanPtr->flags |= TCL_BUFFER_READY;
    }
    
    /*
     * If the flush fails we cannot recover the original position, so
     * do not attempt this, and instead return the error. FlushChannel
     * should have called Tcl_SetErrno() to report the error upwards.
     */
    
    if (FlushChannel(chanPtr, 0) != 0) {
        curPos = -1;
    }
    DiscardInputQueued(chanPtr);
    
    /*
     * Seek back to the logical read or write position, then apply the seek
     * that the user requested.
     */

    newPos = -1;
    if (curPos != -1) {
        newPos = (chanPtr->typePtr->seekProc) (chanPtr->instanceData,
                chanPtr->inFile, chanPtr->outFile, (long) curPos, SEEK_SET,
                &result);
        if (newPos == -1) {
            Tcl_SetErrno(result);
        } else {
            newPos = (chanPtr->typePtr->seekProc) (chanPtr->instanceData,
                    chanPtr->inFile, chanPtr->outFile, (long) offset,
                    mode, &result);
            if (newPos == -1) {
                Tcl_SetErrno(result);
            }
        }
    }
    
    /*
     * Restore to nonblocking mode if that was the previous behavior.
     *
     * NOTE: Even if there was an async flush active we do not restore
     * it now because we already flushed all the queued output, above.
     */
    
    if (wasAsync) {
        chanPtr->flags |= TCL_ASYNC_CHANNEL;
        result = (chanPtr->typePtr->blockModeProc) (chanPtr->instanceData,
                chanPtr->inFile, chanPtr->outFile, TCL_MODE_NONBLOCKING);
        if (result != 0) {
            Tcl_SetErrno(result);
            return -1;
        }
    }
    
    /*
     * If we are seeking to a different location in the channel,
     * clear the bit that denotes a CR in the input, because such
     * saved information will be invalidated.
     */

    if ((mode != SEEK_CUR) || (offset != 0)) {
        chanPtr->flags &= (~(TCL_INPUT_SAW_CR));
    }
    
    return newPos;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_Tell --
 *
 *	Returns the position of the next character to be read/written on
 *	this channel.
 *
 * Results:
 *	A positive integer on success, -1 on failure. If failed,
 *	use Tcl_GetErrno() to retrieve the POSIX error code for the
 *	error that occurred.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_Tell(chan)
    Tcl_Channel chan;			/* The channel to return pos for. */
{
    return Tcl_Seek(chan, 0, SEEK_CUR);
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_Close --
 *
 *	Implements closing of Tcl channels. This is a public function
 *	so that other C facilities can be implemented on top of it.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May close the physical channel if this is the last reference.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_Close(chan)
    Tcl_Channel chan;		/* The channel being closed. */
{
    ChannelHandler *hPtr, *hNext;
				/* Iterate over channel handlers. */
    EventScriptRecord *ePtr, *eNextPtr;
    				/* Iterate over event script records. */
    TclChannel *chanPtr;	/* The real IO channel. */
    int result;			/* Of calling FlushChannel. */

    chanPtr = (TclChannel *) chan;

    if (chanPtr->refCount > 0) {
        panic("called Tcl_Close on channel with refcount > 0");
    }

    /*
     * Remove any scheduled channel handlers.
     */

    if (chanPtr->channelHandlers != (ChannelHandler *) NULL) {
        
        /*
         * Remove all the channel handler records attached to the channel
         * itself.
         */
        
        for (hPtr = chanPtr->channelHandlers;
                 hPtr != (ChannelHandler *) NULL;
                 hPtr = hNext) {
            hNext = hPtr->nextPtr;
            ckfree((char *) hPtr);

            /*
             * NOTE: It is the responsibility of the caller to delete
             * the clientData for this channel handler.
             */
        }
        chanPtr->channelHandlers = (ChannelHandler *) NULL;

        /*
         * Must set the interest mask now to 0, otherwise infinite loops
         * will occur if Tcl_DoOneEvent is called before the channel is
         * finally deleted in FlushChannel (as can happen if an async
         * flush is active for the channel).
         */
        
        chanPtr->interestMask = 0;

        /*
         * Remove any EventScript records for this channel.
         */

        for (ePtr = chanPtr->scriptRecordPtr;
             	ePtr != (EventScriptRecord *) NULL;
                ePtr = eNextPtr) {
            eNextPtr = ePtr->nextPtr;
            ckfree(ePtr->script);
            ckfree((char *) ePtr);
        }
        chanPtr->scriptRecordPtr = (EventScriptRecord *) NULL;
    }

    /*
     * And remove any events for this channel from the event queue.
     */

    Tcl_DeleteEvents(ChannelEventsDeleteProc, (ClientData) chanPtr);

    /*
     * Ensure that the last output buffer will be flushed.
     */
    
    if ((chanPtr->curOutPtr != (ChannelBuffer *) NULL) &&
            (chanPtr->curOutPtr->nextAdded > 0)) {
        chanPtr->flags |= TCL_BUFFER_READY;
    }

    chanPtr->flags |= TCL_CHANNEL_CLOSED;
    
    /*
     * The call to FlushChannel will flush any queued output and invoke
     * the close function of the channel driver, or it will set up the
     * channel to be flushed and closed asynchronously.
     */
    
    result = FlushChannel(chanPtr, 0);
    if (result != 0) {
        return TCL_ERROR;
    }

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_Eof --
 *
 *	Returns 1 if the channel is at EOF, 0 otherwise.
 *
 * Results:
 *	1 or 0, always.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_Eof(chan)
    Tcl_Channel chan;			/* Does this channel have EOF? */
{
    TclChannel *chanPtr;		/* The real channel structure. */

    chanPtr = (TclChannel *) chan;
    return (chanPtr->flags & TCL_CHANNEL_EOF) ? 1 : 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_InputBlocked --
 *
 *	Returns 1 if input is blocked on this channel, 0 otherwise.
 *
 * Results:
 *	0 or 1, always.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_InputBlocked(chan)
    Tcl_Channel chan;			/* Is this channel blocked? */
{
    TclChannel *chanPtr;		/* The real channel structure. */

    chanPtr = (TclChannel *) chan;
    return (chanPtr->flags & TCL_CHANNEL_BLOCKED) ? 1 : 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_GetChannelOption --
 *
 *	Gets a mode associated with an IO channel.
 *
 * Results:
 *	A string denoting whether a particular option is set.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

char *
Tcl_GetChannelOption(chan, optionName)
    Tcl_Channel chan;		/* Channel on which to get mode. */
    char *optionName;		/* Mode to get. */
{
    TclChannel *chanPtr;		/* The real IO channel. */
    size_t len;			/* Length of optionName string. */

    chanPtr = (TclChannel *) chan;

    if (optionName[0] == '-') {
        optionName++;
    } else {
        Tcl_SetErrno(EINVAL);
        return NULL;
    }
    
    len = strlen(optionName);
    
    if ((optionName[0] == 'b') &&
            (strncmp(optionName, "blocking", len) == 0)) {
        return (chanPtr->flags & TCL_ASYNC_CHANNEL) ? "0" : "1";
    }
    if ((optionName[0] == 'l') &&
            (strncmp(optionName, "linemode", len) == 0)) {
        return (chanPtr->flags & TCL_LINE_BUFFERED) ? "1" : "0";
    }
    if ((optionName[0] == 't') &&
            (strncmp(optionName, "translation", len) == 0)) {
        if (chanPtr->translation == TCL_TRANSLATE_AUTO) {
            return "auto";
        }
        if (chanPtr->translation == TCL_TRANSLATE_CR) {
            return "cr";
        }
        if (chanPtr->translation == TCL_TRANSLATE_CRLF) {
            return "crlf";
        }
        if (chanPtr->translation == TCL_TRANSLATE_LF) {
            return "lf";
        }
        panic("Tcl_GetChannelOption: no translation mode set");
    }

    Tcl_SetErrno(EINVAL);
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_SetChannelOption --
 *
 *	Sets an option on a channel.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May modify an option on a device.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_SetChannelOption(interp, chan, optionName, newValue)
    Tcl_Interp *interp;		/* For error reporting - can be NULL. */
    Tcl_Channel chan;		/* Channel on which to set mode. */
    char *optionName;		/* Which option to set? */
    char *newValue;		/* New value for option. */
{
    int result;			/* Result of channel type operation. */
    int newMode;		/* New (numeric) mode to sert. */
    TclChannel *chanPtr;	/* The real IO channel. */
    size_t len;			/* Length of optionName string. */

    chanPtr = (TclChannel *) chan;
    
    len = strlen(optionName);
    if ((optionName[0] != '-') || (len < 2)) {
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "malformed option: ", optionName,
                    (char *) NULL);
        }
        return TCL_ERROR;
    }
    
    if ((optionName[1] == 'b') &&
            (strncmp(optionName, "-blocking", len) == 0)) {
        if (Tcl_GetBoolean(interp, newValue, &newMode) == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (newMode) {
            newMode = TCL_MODE_BLOCKING;
        } else {
            newMode = TCL_MODE_NONBLOCKING;
        }
        result = (chanPtr->typePtr->blockModeProc) (chanPtr->instanceData,
                chanPtr->inFile, chanPtr->outFile, newMode);
        if (result != 0) {
            Tcl_SetErrno(result);
            if (interp != (Tcl_Interp *) NULL) {
                Tcl_AppendResult(interp, "error setting blocking mode: ",
                        Tcl_PosixError(interp), (char *) NULL);
            }
            return TCL_ERROR;
        }
        if (newMode == TCL_MODE_BLOCKING) {
            chanPtr->flags &= (~(TCL_ASYNC_CHANNEL));
            if (chanPtr->outFile != (Tcl_File) NULL) {
                Tcl_DeleteFileHandler(chanPtr->outFile);
                chanPtr->flags &= (~(TCL_ASYNC_FLUSH_ACTIVE));
                if ((chanPtr->curOutPtr != (ChannelBuffer *) NULL) &&
                        (chanPtr->curOutPtr->nextAdded != 0)) {
                    chanPtr->flags |= TCL_BUFFER_READY;
                }
                result = FlushChannel(chanPtr, 0);
                if (result != 0) {
                    if (interp != (Tcl_Interp *) NULL) {
                        Tcl_AppendResult(interp, "error flushing : ",
                                Tcl_PosixError(interp), (char *) NULL);
                    }
                    return TCL_ERROR;
                }
            }
        } else {
            chanPtr->flags |= TCL_ASYNC_CHANNEL;
        }
        return TCL_OK;
    }
    
    if ((optionName[1] == 'l') &&
            (strncmp(optionName, "-linemode", len) == 0)) {
        if (Tcl_GetBoolean(interp, newValue, &newMode) == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (newMode) {
            chanPtr->flags |= TCL_LINE_BUFFERED;
        } else {
            chanPtr->flags &= (~(TCL_LINE_BUFFERED));
        }
        return TCL_OK;
    }

    if ((optionName[1] == 't') &&
            (strncmp(optionName, "-translation", len) == 0)) {
        if (strcmp(newValue, "auto") == 0) {
            chanPtr->translation = TCL_TRANSLATE_AUTO;
        } else if (strcmp(newValue, "lf") == 0) {
            chanPtr->translation = TCL_TRANSLATE_LF;
        } else if (strcmp(newValue, "cr") == 0) {
            chanPtr->flags &= (~(TCL_INPUT_SAW_CR));
            chanPtr->translation = TCL_TRANSLATE_CR;
        } else if (strcmp(newValue, "crlf") == 0) {
            chanPtr->translation = TCL_TRANSLATE_CRLF;
        } else if (strcmp(newValue, "platform") == 0) {
            chanPtr->translation = TCL_PLATFORM_TRANSLATION;
        } else {
            if (interp != (Tcl_Interp *) NULL) {
                Tcl_AppendResult(interp, "illegal value for -translation: ",
                        "must be one of \"cr\", \"lf\", \"crlf\", \"auto\"",
                        " or \"platform\"", (char *) NULL);
            }
            return TCL_ERROR;
        }
        return TCL_OK;
    }

    if (interp != (Tcl_Interp *) NULL) {
        Tcl_AppendResult(interp, "bad option \"", optionName,
                "\": should be -blocking, -linemode or -translation",
                (char *) NULL);
    }

    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * CHANNEL EVENT SOURCE
 *
 *----------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------
 *
 * ChannelEventDispatcher --
 *
 *	This routine is the dispatcher for a channel event. It causes the
 *	consumption of input and production of output on channels during
 *	an event.
 *
 *	Errors that occur during the read or write operation are stored
 *	inside the channel structure for future reporting by the next
 *	operation that uses this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Causes consumption of input and/or production of output on a
 *	channel.
 *
 *----------------------------------------------------------------------
 */

static void
ChannelEventDispatcher(clientData, mask)
    ClientData clientData;		/* The Tcl_Channel to consume
                                         * input from and/or produce output
                                         * on. */
    int mask;				/* ORed combination of TCL_READABLE
                                         * and TCL_WRITABLE. Controls whether
                                         * the dispatcher will cause input to
                                         * be consumed and/or output to be
                                         * produced. */
{
    TclChannel *chanPtr;

    chanPtr = (TclChannel *) clientData;
    if ((mask & (TCL_WRITABLE | TCL_EXCEPTION)) &&
            (chanPtr->outFile != (Tcl_File) NULL)) {
        (void) FlushChannel(chanPtr, 1);
    }
    if ((mask & (TCL_READABLE | TCL_EXCEPTION)) &&
            (chanPtr->inFile != (Tcl_File) NULL)) {
        (void) GetInputBuffer(chanPtr, -1);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * ChannelEventsDeleteProc --
 *
 *	This procedure returns 1 if the event passed in is for the
 *	channel passed in as the second argument.
 *
 * Results:
 *	1 if matching, 0 otherwise.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
ChannelEventsDeleteProc(evPtr, clientData)
    Tcl_Event *evPtr;		/* The event to check for a match. */
    ClientData clientData;	/* The channel to check for. */
{
    ChannelHandlerEvent *cEvPtr;
    TclChannel *chanPtr;

    if (evPtr->proc != ChannelHandlerEventProc) {
        return 0;
    }
    cEvPtr = (ChannelHandlerEvent *) evPtr;
    chanPtr = (TclChannel *) clientData;
    if (cEvPtr->chanPtr != chanPtr) {
        return 0;
    }
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * ChannelHandlerExitProc --
 *
 *	This procedure is called during exit cleanup to delete the channel
 *	event source.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Destroys the channel event source.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
ChannelHandlerExitProc(clientData)
    ClientData clientData;		/* Not used. */
{
    /*
     * Delete the event source for channels.
     */
    
    Tcl_DeleteEventSource(ChannelHandlerSetupProc, ChannelHandlerCheckProc,
            (ClientData) NULL);
}

/*
 *----------------------------------------------------------------------
 *
 * ChannelHandlerSetupProc --
 *
 *	This procedure is part of the "event source" for channel handlers.
 *	It is invoked by Tcl_DoOneEvent before it waits for events. The
 *	job of this procedure is to provide information to Tcl_DoOneEvent
 *	on how to wait for events (what channels to watch).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Tells the notifier what channels to watch.
 *
 *----------------------------------------------------------------------
 */

static void
ChannelHandlerSetupProc(clientData, flags)
    ClientData clientData;		/* Not used. */
    int flags;				/* Flags passed to Tk_DoOneEvent:
					 * if it doesn't include
					 * TCL_FILE_EVENTS then we do
					 * nothing. */
{
    Tcl_Time t;
    TclChannel *chanPtr;
    TclChannel *nextChanPtr;

    if (!(flags & TCL_FILE_EVENTS)) {
        return;
    }

    t.sec = 0; t.usec = 0;
    
    for (chanPtr = firstChanPtr; chanPtr != (TclChannel *) NULL;
             chanPtr = nextChanPtr) {
        nextChanPtr = chanPtr->nextChanPtr;
        if (chanPtr->interestMask & TCL_READABLE) {
            if (chanPtr->inQueueHead != (ChannelBuffer *) NULL) {
                Tcl_SetMaxBlockTime(&t);
            } else if (chanPtr->inFile != (Tcl_File) NULL) {
                Tcl_WatchFile(chanPtr->inFile, TCL_READABLE);
            }
        }
        if (chanPtr->interestMask & TCL_WRITABLE) {
            if (chanPtr->outFile != (Tcl_File) NULL) {
                Tcl_WatchFile(chanPtr->outFile, TCL_WRITABLE);
            }
        }
        if (chanPtr->interestMask & TCL_EXCEPTION) {
            if (chanPtr->inFile != (Tcl_File) NULL) {
                Tcl_WatchFile(chanPtr->inFile, TCL_EXCEPTION);
            }
            if (chanPtr->outFile != (Tcl_File) NULL) {
                Tcl_WatchFile(chanPtr->outFile, TCL_EXCEPTION);
            }
        }
    }
}

/*
 *----------------------------------------------------------------------
 *
 * ChannelHandlerEventProc --
 *
 *	This procedure is called by Tcl_DoOneEvent when a channel event
 *	reaches the front of the event queue. This procedure is responsible
 *	for actually handling the event by invoking the callback for the
 *	channel handler.
 *
 * Results:
 *	Returns 1 if the event was handled, meaning that it should be
 *	removed from the queue. Returns 0 if the event was not handled
 *	meaning that it should stay in the queue. The only time the event
 *	will not be handled is if the TCL_FILE_EVENTS flag bit is not
 *	set in the flags passed.
 *
 *	NOTE: If the handler is deleted between the time the event is added
 *	to the queue and the time it reaches the head of the queue, the
 *	event is silently discarded (i.e. we return 1).
 *
 * Side effects:
 *	Whatever the channel handler callback procedure does.
 *
 *----------------------------------------------------------------------
 */

static int
ChannelHandlerEventProc(evPtr, flags)
    Tcl_Event *evPtr;		/* Event to service. */
    int flags;			/* Flags that indicate what events to
                                 * handle, such as TCL_FILE_EVENTS. */
{
    TclChannel *chanPtr;
    ChannelHandler *hPtr, *nextPtr;
    ChannelHandlerEvent *ePtr;

    if (!(flags & TCL_FILE_EVENTS)) {
        return 0;
    }

    ePtr = (ChannelHandlerEvent *) evPtr;
    chanPtr = ePtr->chanPtr;

    for (hPtr = chanPtr->channelHandlers;
             hPtr != (ChannelHandler *) NULL;
             hPtr = nextPtr) {

        nextPtr = hPtr->nextPtr;

        /*
         * If this channel handler is interested in any of the events that
         * have occurred on the channel, invoke its procedure.
         */
        
        if ((hPtr->mask & ePtr->readyMask) != 0) {
	    (*hPtr->proc) (hPtr->clientData, ePtr->readyMask);
	}
    }
    
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * ChannelHandlerCheckProc --
 *
 *	This procedure is the second part (of three) of the "event source"
 *	for channels. It is invoked by Tcl_DoOneEvent after the wait for
 *	events is over. The job of this procedure is to test each channel
 *	to see if it is ready now, and if so, to create events and put them
 *	on the Tcl event queue.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Makes entries on the Tcl event queue for each channel that is
 *	ready now.
 *
 *----------------------------------------------------------------------
 */

static void
ChannelHandlerCheckProc(clientData, flags)
    ClientData clientData;		/* Not used. */
    int flags;				/* Flags passed to Tk_DoOneEvent:
					 * if it doesn't include 
					 * TCL_FILE_EVENTS then we do
					 * nothing. */
{
    TclChannel *chanPtr;
    TclChannel *nextChanPtr;
    ChannelHandlerEvent *ePtr;
    int readyMask;
    
    if (!(flags & TCL_FILE_EVENTS)) {
        return;
    }

    for (chanPtr = firstChanPtr;
             chanPtr != (TclChannel *) NULL;
             chanPtr = nextChanPtr) {
        nextChanPtr = chanPtr->nextChanPtr;

        readyMask = 0;

        /*
         * Check for readability.
         */
        
        if (chanPtr->interestMask & TCL_READABLE) {

            /*
             * The channel is considered ready for reading if there is input
             * buffered AND the channel is not blocked, OR if the underlying
             * file is ready.
             */
            
            if ((!(chanPtr->flags & TCL_CHANNEL_BLOCKED)) &&
                    (chanPtr->inQueueHead != (ChannelBuffer *) NULL)) {
                readyMask |= TCL_READABLE;
            } else if (chanPtr->inFile != (Tcl_File) NULL) {
                readyMask |=
                    Tcl_FileReady(chanPtr->inFile, TCL_READABLE);
            }
        }

        /*
         * Check for writability.
         */

        if (chanPtr->interestMask & TCL_WRITABLE) {

            /*
             * The channel is considered ready for writing if there is no
             * output buffered AND the underlying file is ready.
             */
            
            if ((chanPtr->outQueueHead == (ChannelBuffer *) NULL) &&
                    (chanPtr->outFile != (Tcl_File) NULL)) {
                readyMask |=
                    Tcl_FileReady(chanPtr->outFile, TCL_WRITABLE);
            }
        }

        /*
         * Check for exceptions.
         */

        if (chanPtr->interestMask & TCL_EXCEPTION) {
            if (chanPtr->inFile != (Tcl_File) NULL) {
                readyMask |=
                    Tcl_FileReady(chanPtr->inFile, TCL_EXCEPTION);
            }
            if (chanPtr->outFile != (Tcl_File) NULL) {
                readyMask |=
                    Tcl_FileReady(chanPtr->outFile, TCL_EXCEPTION);
            }
        }
        
        /*
         * If there are any events for this channel, put a notice into the
         * Tcl event queue.
         */
        
        if (readyMask != 0) {
            ePtr = (ChannelHandlerEvent *) ckalloc((unsigned)
                    sizeof(ChannelHandlerEvent));
            ePtr->header.proc = ChannelHandlerEventProc;
            ePtr->chanPtr = chanPtr;
            ePtr->readyMask = readyMask;
            Tcl_QueueEvent((Tcl_Event *) ePtr, TCL_QUEUE_TAIL);
        }
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_CreateChannelHandler --
 *
 *	Arrange for a given procedure to be invoked whenever the
 *	channel indicated by the chanPtr arg becomes readable or
 *	writable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	From now on, whenever the I/O channel given by chanPtr becomes
 *	ready in the way indicated by mask, proc will be invoked.
 *	See the manual entry for details on the calling sequence
 *	to proc.  If the channel is already registered then the old mask,
 *	proc and clientData values will be replaced with new ones.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_CreateChannelHandler(chan, mask, proc, clientData)
    Tcl_Channel chan;		/* The channel to create the handler for. */
    int mask;			/* OR'ed combination of TCL_READABLE,
				 * TCL_WRITABLE, and TCL_EXCEPTION:
				 * indicates conditions under which
				 * proc should be called. Use 0 to
                                 * disable a registered handler. */
    Tcl_FileProc *proc;		/* Procedure to call for each
				 * selected event. */
    ClientData clientData;	/* Arbitrary data to pass to proc. */
{
    ChannelHandler *hPtr;
    TclChannel *chanPtr;

    chanPtr = (TclChannel *) chan;
    
    /*
     * Ensure that the channel event source is registered with the Tcl
     * notification mechanism.
     */
    
    if (!channelEventSourceCreated) {
        channelEventSourceCreated = 1;
        Tcl_CreateEventSource(ChannelHandlerSetupProc, ChannelHandlerCheckProc,
                (ClientData) NULL);
        Tcl_CreateExitHandler(ChannelHandlerExitProc, (ClientData) NULL);
    }

    /*
     * Check whether this channel handler is not already registered. If
     * it is not, create a new record, else reuse existing record (smash
     * current values).
     */

    for (hPtr = chanPtr->channelHandlers;
             hPtr != (ChannelHandler *) NULL;
             hPtr = hPtr->nextPtr) {
        if ((hPtr->chanPtr == chanPtr) && (hPtr->proc == proc) &&
                (hPtr->clientData == clientData)) {
            break;
        }
    }
    if (hPtr == (ChannelHandler *) NULL) {
        hPtr = (ChannelHandler *) ckalloc((unsigned) sizeof(ChannelHandler));
        hPtr->mask = 0;
        hPtr->chanPtr = chanPtr;
        hPtr->nextPtr = chanPtr->channelHandlers;
        chanPtr->channelHandlers = hPtr;
    }

    /*
     * The remainder of the initialization below is done regardless of
     * whether or not this is a new record or a modification of an old
     * one.
     */

    hPtr->mask = mask;
    hPtr->proc = proc;
    hPtr->clientData = clientData;

    /*
     * Recompute the interest mask for the channel - this call may actually
     * be disabling an existing handler..
     */
    
    chanPtr->interestMask = 0;
    for (hPtr = chanPtr->channelHandlers;
	     hPtr != (ChannelHandler *) NULL;
	     hPtr = hPtr->nextPtr) {
	chanPtr->interestMask |= hPtr->mask;
    }                                       
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_DeleteChannelHandler --
 *
 *	Cancel a previously arranged callback arrangement for an IO
 *	channel. Also removes any scripts stored for this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	If a callback was previously registered for this channel, it
 *	is removed and the callback will no longer be called when the
 *	channel becomes ready for IO.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_DeleteChannelHandler(chan, proc, clientData)
    Tcl_Channel chan;		/* The channel for which to remove the
                                 * callback. */
    Tcl_FileProc *proc;		/* The procedure in the callback to delete. */
    ClientData clientData;	/* The client data in the callback
                                 * to delete. */
    
{
    ChannelHandler *hPtr, *hPrev;
    TclChannel *chanPtr;

    chanPtr = (TclChannel *) chan;

    /*
     * Find the entry and the previous one in the list.
     */

    for (hPrev = (ChannelHandler *) NULL, hPtr = chanPtr->channelHandlers;
             hPtr != (ChannelHandler *) NULL;
             hPtr = hPtr->nextPtr) {
        if ((hPtr->chanPtr == chanPtr) && (hPtr->clientData == clientData)
                && (hPtr->proc == proc)) {
            break;
        }
        hPrev = hPtr;
    }

    /*
     * If found, splice the entry out of the list.
     */

    if (hPtr == (ChannelHandler *) NULL) {
        return;
    }

    if (hPrev == (ChannelHandler *) NULL) {
        chanPtr->channelHandlers = hPtr->nextPtr;
    } else {
        hPrev->nextPtr = hPtr->nextPtr;
    }
    ckfree((char *) hPtr);

    /*
     * Recompute the interest list for the channel, so that infinite loops
     * will not result if Tcl_DeleteChanelHandler is called inside an event.
     */

    chanPtr->interestMask = 0;
    for (hPtr = chanPtr->channelHandlers;
             hPtr != (ChannelHandler *) NULL;
             hPtr = hPtr->nextPtr) {
        chanPtr->interestMask |= hPtr->mask;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * ReturnScriptRecord --
 *
 *	Get a script stored for this channel with this interpreter.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Sets interp->result to the script.
 *
 *----------------------------------------------------------------------
 */

static void
ReturnScriptRecord(interp, chanPtr, mask)
    Tcl_Interp *interp;		/* The interpreter in which the script
                                 * is to be executed. */
    TclChannel *chanPtr;	/* The channel for which the script is
                                 * stored. */
    int mask;			/* Events in mask must overlap with events
                                 * for which this script is stored. */
{
    EventScriptRecord *sPtr;
    
    for (sPtr = chanPtr->scriptRecordPtr;
             sPtr != (EventScriptRecord *) NULL;
             sPtr = sPtr->nextPtr) {
        if ((sPtr->interp == interp) && (sPtr->mask == mask)) {
            Tcl_AppendResult(interp, sPtr->script, (char *) NULL);
            return;
        }
    }
}

/*
 *----------------------------------------------------------------------
 *
 * DeleteScriptRecord --
 *
 *	Delete a script record for this combination of channel, interp
 *	and mask.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Deletes a script record and cancels a channel event handler.
 *
 *----------------------------------------------------------------------
 */

static void
DeleteScriptRecord(interp, chanPtr, mask)
    Tcl_Interp *interp;		/* Interpreter in which script was to be
                                 * executed. */
    TclChannel *chanPtr;		/* The channel for which to delete the
                                 * script record (if any). */
    int mask;			/* Events in mask must exactly match mask
                                 * of script to delete. */
{
    EventScriptRecord *sPtr, *nextPtr, *prevPtr;

    for (sPtr = chanPtr->scriptRecordPtr, prevPtr = (EventScriptRecord *) NULL;
             sPtr != (EventScriptRecord *) NULL;
             sPtr = nextPtr) {
        nextPtr = sPtr->nextPtr;
        if ((sPtr->interp == interp) && (sPtr->mask == mask)) {
            if (sPtr == chanPtr->scriptRecordPtr) {
                chanPtr->scriptRecordPtr = sPtr->nextPtr;
            } else {
                prevPtr->nextPtr = sPtr->nextPtr;
            }

            Tcl_DeleteChannelHandler((Tcl_Channel) chanPtr,
                    ChannelEventScriptInvoker, (ClientData) sPtr);
            
            ckfree(sPtr->script);
            ckfree((char *) sPtr);

            break;
        } else {
            prevPtr = sPtr;
        }
    }
}

/*
 *----------------------------------------------------------------------
 *
 * CreateScriptRecord --
 *
 *	Creates a record to store a script to be executed when a specific
 *	event fires on a specific channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Causes the script to be stored for later execution.
 *
 *----------------------------------------------------------------------
 */

static void
CreateScriptRecord(interp, chanPtr, mask, script)
    Tcl_Interp *interp;			/* Interpreter in which to execute
                                         * the stored script. */
    TclChannel *chanPtr;		/* Channel for which script is to
                                         * be stored. */
    int mask;				/* Set of events for which script
                                         * will be invoked. */
    char *script;			/* A copy of this script is stored
                                         * in the newly created record. */
{
    EventScriptRecord *sPtr;

    for (sPtr = chanPtr->scriptRecordPtr;
             sPtr != (EventScriptRecord *) NULL;
             sPtr = sPtr->nextPtr) {
        if ((sPtr->interp == interp) && (sPtr->mask == mask)) {
            ckfree(sPtr->script);
            sPtr->script = (char *) NULL;
            break;
        }
    }
    if (sPtr == (EventScriptRecord *) NULL) {
        sPtr = (EventScriptRecord *) ckalloc((unsigned)
                sizeof(EventScriptRecord));
        Tcl_CreateChannelHandler((Tcl_Channel) chanPtr, mask,
                ChannelEventScriptInvoker, (ClientData) sPtr);
        sPtr->nextPtr = chanPtr->scriptRecordPtr;
        chanPtr->scriptRecordPtr = sPtr;
    }
    sPtr->chan = (Tcl_Channel) chanPtr;
    sPtr->interp = interp;
    sPtr->mask = mask;
    sPtr->script = ckalloc((unsigned) strlen(script) + 1);
    strcpy(sPtr->script, script);
}

/*
 *----------------------------------------------------------------------
 *
 * ChannelEventScriptInvoker --
 *
 *	Invokes a script scheduled by "fileevent" for when the channel
 *	becomes ready for IO.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Whatever the scripts do.
 *
 *----------------------------------------------------------------------
 */

static void
ChannelEventScriptInvoker(clientData, mask)
    ClientData clientData;	/* The script+interp record. */
    int mask;			/* Not used. */
{
    Tcl_Interp *interp;		/* Interpreter in which to eval the script. */
    TclChannel *chanPtr;	/* The channel for which this handler is
                                 * registered. */
    Tcl_DString script;		/* Copy of script to eval. */
    EventScriptRecord *sPtr;	/* The event script + interpreter to eval it
                                 * in. */
    EventScriptRecord *sNext, *sPrev;
    				/* Iterate over list of script records. */
    int result;			/* Result of call to eval script. */

    sPtr = (EventScriptRecord *) clientData;

    chanPtr = (TclChannel *) sPtr->chan;

    mask = sPtr->mask;
    interp = sPtr->interp;
    
    Tcl_DStringInit(&script);
    Tcl_DStringAppend(&script, sPtr->script, -1);
            
    result = Tcl_GlobalEval(sPtr->interp, Tcl_DStringValue(&script));

    Tcl_DStringFree(&script);

    /*
     * On error, cause a background error and remove the channel handler
     * and the script record if they have not yet been removed. We must
     * search and find the record in the global list because sPtr may
     * not be valid anymore after the evaluation.
     */
    
    if (result != TCL_OK) {
        Tcl_BackgroundError(interp);

        for (sPtr = chanPtr->scriptRecordPtr,
                 sPrev = (EventScriptRecord *) NULL;
             sPtr != (EventScriptRecord *) NULL;
             sPtr = sNext) {
            sNext = sPtr->nextPtr;
            if ((sPtr->mask == mask) && (sPtr->interp == interp)) {

                Tcl_DeleteChannelHandler((Tcl_Channel) chanPtr,
                        ChannelEventScriptInvoker, (ClientData) sPtr);

                ckfree(sPtr->script);
                ckfree((char *) sPtr);

                if (sPrev == (EventScriptRecord *) NULL) {
                    chanPtr->scriptRecordPtr = sNext;
                } else {
                    sPrev->nextPtr = sNext;
                }

                break;
            } else {
                sPrev = sPtr;
            }
        }
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_ChannelEventCmd --
 *
 *	This procedure implements the "fileevent" Tcl command. See the
 *	user documentation for details on what it does. This command is
 *	based on the Tk command "fileevent" which in turn is based on work
 *	contributed by Mark Diekhans.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May create a channel handler for the specified channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_ChannelEventCmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Not used. */
    Tcl_Interp *interp;			/* Interpreter in which the channel
                                         * for which to create the handler
                                         * is found. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    TclChannel *chanPtr;			/* The channel to create
                                         * the handler for. */
    Tcl_Channel chan;			/* The opaque type for the channel. */
    int c;				/* First char of mode argument. */
    int mask;				/* Mask for events of interest. */
    size_t length;			/* Length of mode argument. */

    /*
     * Parse arguments.
     */

    if ((argc != 3) && (argc != 4)) {
	Tcl_AppendResult(interp, "wrong # args: must be \"", argv[0],
		" channelId event ?script?", (char *) NULL);
	return TCL_ERROR;
    }
    c = argv[2][0];
    length = strlen(argv[2]);
    if ((c == 'r') && (strncmp(argv[2], "readable", length) == 0)) {
        mask = TCL_READABLE;
    } else if ((c == 'w') && (strncmp(argv[2], "writable", length) == 0)) {
        mask = TCL_WRITABLE;
    } else {
	Tcl_AppendResult(interp, "bad event name \"", argv[2],
		"\": must be readable or writable", (char *) NULL);
	return TCL_ERROR;
    }
    chan = Tcl_GetChannel(interp, argv[1], NULL);
    if (chan == (Tcl_Channel) NULL) {
        return TCL_ERROR;
    }
    
    chanPtr = (TclChannel *) chan;
    if ((chanPtr->flags & mask) == 0) {
        Tcl_AppendResult(interp, "\"", argv[1], "\": wrong mode",
                (char *) NULL);
        return TCL_ERROR;
    }
    
    /*
     * If we are supposed to return the script (argc == 3), do so.
     */

    if (argc == 3) {
        ReturnScriptRecord(interp, chanPtr, mask);
        return TCL_OK;
    }

    /*
     * If we are supposed to delete a stored script, do so.
     */

    if ((argv[3] == (char *) NULL) || (argv[3][0] == 0)) {
        DeleteScriptRecord(interp, chanPtr, mask);
        return TCL_OK;
    }

    /*
     * Make the script record that will link between the event and the
     * script to invoke. This also creates a channel event handler which
     * will evaluate the script in the supplied interpreter.
     */

    CreateScriptRecord(interp, chanPtr, mask, argv[3]);
    
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TclTestChannelCmd --
 *
 *	Implements the Tcl "testchannel" debugging command and its
 *	subcommands. This is part of the testing environment but must be
 *	in this file instead of tclTest.c because it needs access to the
 *	fields of struct TclChannel.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
TclTestChannelCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Interpreter for result. */
    int argc;			/* Count of additional args. */
    char **argv;		/* Additional arg strings. */
{
    char *cmdName;		/* Sub command. */
    Tcl_HashTable *hTblPtr;	/* Hash table of channels. */
    Tcl_HashSearch hSearch;	/* Search variable. */
    Tcl_HashEntry *hPtr;	/* Search variable. */
    TclChannel *chanPtr;	/* The actual channel. */
    Tcl_Channel chan;		/* The opaque type. */
    size_t len;			/* Length of subcommand string. */
    int IOQueued;		/* How much IO is queued inside channel? */
    ChannelBuffer *run;		/* For iterating over queued IO. */
    char buf[128];		/* For sprintf. */
    
    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                " subcommand ?additional args..?\"", (char *) NULL);
        return TCL_ERROR;
    }
    cmdName = argv[1];
    len = strlen(cmdName);

    if ((cmdName[0] == 'i') && (strncmp(cmdName, "info", len) == 0)) {
        if (argc != 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                    " info channelname\"", (char *) NULL);
            return TCL_ERROR;
        }
        chan = Tcl_GetChannel(interp, argv[2], NULL);
        if (chan == (Tcl_Channel) NULL) {
            return TCL_ERROR;
        }
        chanPtr = (TclChannel *) chan;
        Tcl_AppendElement(interp, argv[2]);
        Tcl_AppendElement(interp, chanPtr->typePtr->typeName);
        if (chanPtr->flags & TCL_READABLE) {
            Tcl_AppendElement(interp, "read");
        } else {
            Tcl_AppendElement(interp, "");
        }
        if (chanPtr->flags & TCL_WRITABLE) {
            Tcl_AppendElement(interp, "write");
        } else {
            Tcl_AppendElement(interp, "");
        }
        if (chanPtr->flags & TCL_ASYNC_CHANNEL) {
            Tcl_AppendElement(interp, "nonblocking");
        } else {
            Tcl_AppendElement(interp, "blocking");
        }
        if (chanPtr->flags & TCL_LINE_BUFFERED) {
            Tcl_AppendElement(interp, "linemode");
        } else {
            Tcl_AppendElement(interp, "");
        }
        if (chanPtr->flags & TCL_ASYNC_FLUSH_ACTIVE) {
            Tcl_AppendElement(interp, "async_flush");
        } else {
            Tcl_AppendElement(interp, "");
        }
        if (chanPtr->flags & TCL_CHANNEL_EOF) {
            Tcl_AppendElement(interp, "eof");
        } else {
            Tcl_AppendElement(interp, "");
        }
        if (chanPtr->flags & TCL_CHANNEL_BLOCKED) {
            Tcl_AppendElement(interp, "blocked");
        } else {
            Tcl_AppendElement(interp, "unblocked");
        }
        if (chanPtr->translation == TCL_TRANSLATE_AUTO) {
            Tcl_AppendElement(interp, "auto");
        } else if (chanPtr->translation == TCL_TRANSLATE_LF) {
            Tcl_AppendElement(interp, "lf");
        } else if (chanPtr->translation == TCL_TRANSLATE_CR) {
            Tcl_AppendElement(interp, "cr");
        } else if (chanPtr->translation == TCL_TRANSLATE_CRLF) {
            Tcl_AppendElement(interp, "crlf");
        }
        for (IOQueued = 0, run = chanPtr->inQueueHead;
                 run != (ChannelBuffer *) NULL;
                 run = run->nextPtr) {
            IOQueued += run->nextAdded - run->nextRemoved;
        }
        sprintf(buf, "%d", IOQueued);
        Tcl_AppendElement(interp, buf);
        
        IOQueued = 0;
        if (chanPtr->curOutPtr != (ChannelBuffer *) NULL) {
            IOQueued = chanPtr->curOutPtr->nextAdded -
                chanPtr->curOutPtr->nextRemoved;
        }
        for (run = chanPtr->outQueueHead;
                 run != (ChannelBuffer *) NULL;
                 run = run->nextPtr) {
            IOQueued += (run->nextAdded - run->nextRemoved);
        }
        sprintf(buf, "%d", IOQueued);
        Tcl_AppendElement(interp, buf);
        
        sprintf(buf, "%d", Tcl_Tell((Tcl_Channel) chanPtr));
        Tcl_AppendElement(interp, buf);

        sprintf(buf, "%d", chanPtr->refCount);
        Tcl_AppendElement(interp, buf);

        return TCL_OK;
    }
    
    if ((cmdName[0] == 'o') && (strncmp(cmdName, "open", len) == 0)) {
        hTblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp, "tclIO", NULL);
        if (hTblPtr == (Tcl_HashTable *) NULL) {
            return TCL_OK;
        }
        for (hPtr = Tcl_FirstHashEntry(hTblPtr, &hSearch);
                 hPtr != (Tcl_HashEntry *) NULL;
                 hPtr = Tcl_NextHashEntry(&hSearch)) {
            Tcl_AppendElement(interp, Tcl_GetHashKey(hTblPtr, hPtr));
        }
        return TCL_OK;
    }

    if ((cmdName[0] == 'r') && (strncmp(cmdName, "readable", len) == 0)) {
        hTblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp, "tclIO", NULL);
        if (hTblPtr == (Tcl_HashTable *) NULL) {
            return TCL_OK;
        }
        for (hPtr = Tcl_FirstHashEntry(hTblPtr, &hSearch);
                 hPtr != (Tcl_HashEntry *) NULL;
                 hPtr = Tcl_NextHashEntry(&hSearch)) {
            chanPtr = (TclChannel *) Tcl_GetHashValue(hPtr);
            if (chanPtr->flags & TCL_READABLE) {
                Tcl_AppendElement(interp, Tcl_GetHashKey(hTblPtr, hPtr));
            }
        }
        return TCL_OK;
    }
    
    if ((cmdName[0] == 'w') && (strncmp(cmdName, "writable", len) == 0)) {
        hTblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp, "tclIO", NULL);
        if (hTblPtr == (Tcl_HashTable *) NULL) {
            return TCL_OK;
        }
        for (hPtr = Tcl_FirstHashEntry(hTblPtr, &hSearch);
                 hPtr != (Tcl_HashEntry *) NULL;
                 hPtr = Tcl_NextHashEntry(&hSearch)) {
            chanPtr = (TclChannel *) Tcl_GetHashValue(hPtr);
            if (chanPtr->flags & TCL_WRITABLE) {
                Tcl_AppendElement(interp, Tcl_GetHashKey(hTblPtr, hPtr));
            }
        }
        return TCL_OK;
    }
    
    if ((cmdName[0] == 'w') && (strncmp(cmdName, "withtype", len) == 0)) {
        if (argc != 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                    " withtype ioTypeName\"", (char *) NULL);
            return TCL_ERROR;
        }
        hTblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp, "tclIO", NULL);
        if (hTblPtr == (Tcl_HashTable *) NULL) {
            return TCL_OK;
        }
        for (hPtr = Tcl_FirstHashEntry(hTblPtr, &hSearch);
                 hPtr != (Tcl_HashEntry *) NULL;
                 hPtr = Tcl_NextHashEntry(&hSearch)) {
            chanPtr = (TclChannel *) Tcl_GetHashValue(hPtr);
            if (strcmp(chanPtr->typePtr->typeName, argv[2]) == 0) {
                Tcl_AppendElement(interp, Tcl_GetHashKey(hTblPtr, hPtr));
            }
        }
        return TCL_OK;
    }

    Tcl_AppendResult(interp, "bad option \"", cmdName, "\": should be ",
            "info, open, readable, withtype or writable",
            (char *) NULL);
    return TCL_ERROR;
}
