/* 
 * tclWinSock.c --
 *
 *	This file contains Windows-specific socket related code.
 *
 * Copyright (c) 1995-1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclWinSock.c 1.11 96/01/30 17:19:44
 */

#include "tclInt.h"
#include "tclPort.h"

/*
 * The preferred buffer size for Windows socket channels.
 */

#define TCP_BUFFER_SIZE 4096

/*
 * The following define declares a new user message for use on the
 * socket window.
 */

#define SOCKET_MESSAGE	WM_USER+1

/*
 * The following structure is used to store the data associated with
 * each socket.  A Tcl_File of type TCL_WIN_SOCKET will contain a
 * pointer to one of these structures in the clientdata slot.
 */

typedef struct SocketInfo {
    SOCKET socket;		   /* Windows SOCKET handle. */
    int flags;			   /* Bit field comprised of the flags
				    * described below.  */
    int checkMask;		   /* OR'ed combination of TCL_READABLE and
				    * TCL_WRITABLE as set by an asynchronous
				    * event handler. */
    int watchMask;		   /* OR'ed combination of TCL_READABLE and
				    * TCL_WRITABLE as set by Tcl_WatchFile. */
    Tcl_File file;		   /* The file handle for the socket. */
    Tcl_TcpAcceptProc *acceptProc; /* Proc to call on accept. */
    ClientData acceptProcData;	   /* The data for the accept proc. */
    struct SocketInfo *nextPtr;	   /* The next socket on the global socket
				    * list. */
} SocketInfo;

/*
 * The following macros may be used to set the flags field of
 * a SocketInfo structure.
 */

#define SOCKET_WATCH 	(1<<1)
				/* TclWinWatchSocket has been called since the
				 * last time we entered Tcl_WaitForEvent. */
#define SOCKET_REGISTERED (1<<2)
				/* A valid WSAAsyncSelect handler is
				 * registered. */
#define SOCKET_ASYNCH	(1<<3)
				/* The socket is in asynch mode. */

/*
 * Every open socket has an entry on the following list.
 */

static SocketInfo *socketList = NULL;

/*
 * Static functions defined in this file.
 */

static void		CleanupSockets _ANSI_ARGS_((ClientData clientData));
static SocketInfo *	CreateSocket _ANSI_ARGS_((Tcl_Interp *interp,
			    int port, char *host, int server));
static int		InitSockets _ANSI_ARGS_((void));
static LRESULT CALLBACK	SocketProc _ANSI_ARGS_((HWND hwnd, UINT message,
			    WPARAM wParam, LPARAM lParam));
static void		TcpAccept _ANSI_ARGS_((ClientData data, int mask));
static int		TcpBlockMode _ANSI_ARGS_((ClientData instanceData,
	            	    Tcl_File inFile, Tcl_File outFile, int mode));
static int		TcpBufferSize _ANSI_ARGS_((ClientData instanceData));
static int		TcpClose _ANSI_ARGS_((ClientData instanceData,
	            	    Tcl_File inFile, Tcl_File outFile));
static int		TcpInput _ANSI_ARGS_((ClientData instanceData,
	            	    Tcl_File inFile, char *buf, int toRead,
	            	    int *errorCode));
static int		TcpOutput _ANSI_ARGS_((ClientData instanceData,
	            	    Tcl_File outFile, char *buf, int toWrite,
	            	    int *errorCode));

/*
 * This structure describes the channel type structure for TCP socket
 * based IO.
 */

static Tcl_ChannelType tcpChannelType = {
    "tcp",		/* Type name. */
    TcpBlockMode,	/* Set blocking or non-blocking mode.*/
    TcpBufferSize,	/* Buffer size. */
    TcpClose,		/* Close proc. */
    TcpInput,		/* Input proc. */
    TcpOutput,		/* Output proc. */
    NULL		/* Seek proc. */
};

/*
 * Remembers that socket initialization was done.
 */

static int              socketsInitialized = 0;
static WSADATA          stWSAData;

/*
 * Socket notification window.  This window is used to receive socket
 * notification events.
 */

static HWND socketWindow = NULL;

/*
 * Window class for creating the socket notification window.
 */

static ATOM socketClass;

/*
 * Highest required version of WinSock:
 */

/* WinSock version information (use WSA_VERSION for WSAStartup()) */
#ifndef MAKEWORD
#define MAKEWORD(l,h) ((WORD)(((BYTE)(l))|(((WORD)(BYTE)(h))<<8)))
#endif
#define WSA_MAJOR_VERSION 1
#define WSA_MINOR_VERSION 1
#define WSA_VERSION MAKEWORD(WSA_MAJOR_VERSION, WSA_MINOR_VERSION)


/*
 *----------------------------------------------------------------------
 *
 * InitSockets --
 *
 *	Initialize the socket module.  Note that WSAStartup was called by the
 *	DllEntryPoint code, so we can assume that the WinSock library has
 *	already been initialized. 
 *
 * Results:
 *	Returns 1 on successful initialization, 0 on failure.
 *
 * Side effects:
 *	Registers a new window class and creates a window for use
 * 	in asynchronous socket notification.
 *
 *----------------------------------------------------------------------
 */

static int
InitSockets()
{
    int status;
    WNDCLASS class;

    /*
     * Note that WSAStartup was called by the DllEntryPoint code, so
     * we can assume that the WinSock library has already been initialized.
     */

    class.style = 0;
    class.cbClsExtra = 0;
    class.cbWndExtra = 0;
    class.hInstance = TclWinGetTclInstance();
    class.hbrBackground = NULL;
    class.lpszMenuName = NULL;
    class.lpszClassName = "TclSocket";
    class.lpfnWndProc = SocketProc;
    class.hIcon = NULL;
    class.hCursor = NULL;

    /*
     * Register the async notification window class and window.
     */

    socketClass = RegisterClass(&class);
    if (!socketClass) {
	TclWinConvertError(GetLastError());
	WSACleanup();
	return 0;
    }
    socketWindow = CreateWindowEx(0, (LPCTSTR)socketClass, "TclSocket",
	    WS_OVERLAPPED, 0, 0, 0, 0, NULL, NULL,
	    TclWinGetTclInstance(), NULL);
    if (socketWindow == NULL) {
	TclWinConvertError(GetLastError());
	UnregisterClass((LPCTSTR)socketClass, TclWinGetTclInstance());
	WSACleanup();
	return 0;
    }
    socketsInitialized = 1;
    Tcl_CreateExitHandler(CleanupSockets, (ClientData) NULL);
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * CleanupSockets --
 *
 *	Callback invoked during exit clean up to release the WinSock
 *      DLL.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Releases the WinSock DLL.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
static void
CleanupSockets(clientData)
    ClientData clientData;              /* Not used. */
{
    DestroyWindow(socketWindow);
    UnregisterClass((LPCTSTR)socketClass, TclWinGetTclInstance());
    WSACleanup();
}

/*
 *----------------------------------------------------------------------
 *
 * TcpClose --
 *
 *	Close the socket.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the socket.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpClose(instanceData, inFile, outFile)
    ClientData instanceData;	/* The socket to close. */
    Tcl_File inFile, outFile;	/* Unused. */
{
    SocketInfo *infoPtr = (SocketInfo *) instanceData;
    int errorCode = 0;

    /*
     * We assume that inFile==outFile==sock and so only clean up one of them.
     */
    
    if (closesocket(infoPtr->socket) == SOCKET_ERROR) {
	TclWinConvertWSAError(WSAGetLastError());
	errorCode = errno;
    }

    /*
     * Delete a file handler that may be active for this socket.
     * Channel handlers are already deleted in the generic IO close
     * code which called this function.
     */
    
    Tcl_DeleteFileHandler(infoPtr->file);

    Tcl_FreeFile(infoPtr->file);

    /*
     * Remove the socket from socketList.
     */

    if (infoPtr == socketList) {
	socketList = infoPtr->nextPtr;
    } else {
	SocketInfo *p;
	for (p = socketList; p != NULL; p = p->nextPtr) {
	    if (p->nextPtr == infoPtr) {
		p->nextPtr = infoPtr->nextPtr;
		break;
	    }
	}
    }
    ckfree((char *) infoPtr);

    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * CreateSocket --
 *
 *	This function opens a new socket and initializes the
 *	SocketInfo structure.
 *
 * Results:
 *	Returns a new SocketInfo, or NULL with an error in interp.
 *
 * Side effects:
 *	Adds a new socket to the socketList.
 *
 *----------------------------------------------------------------------
 */

static SocketInfo *
CreateSocket(interp, port, host, server)
    Tcl_Interp *interp;		/* For error reporting; can be NULL. */
    int port;			/* Port number to open. */
    char *host;			/* Name of host on which to open port. */
    int server;			/* 1 if socket should be a server socket,
				 * else 0 for a client socket. */
{
    int status;
    struct sockaddr_in sockaddr;	/* Socket address */
    struct hostent *hostent;		/* Host database entry */
    struct in_addr addr;		/* For 64/32 bit madness */
    struct linger ling;			/* For making sockets linger. */
    SOCKET sock;
    SocketInfo *infoPtr;

    if (!socketsInitialized) {
	if (!InitSockets()) {
	    Tcl_AppendResult(interp, "couldn't initialize WinSock subsystem",
		    NULL);
	    return NULL;
	}
    }
    (void) memset((char *) &sockaddr, '\0', sizeof(struct sockaddr_in));
    sockaddr.sin_family = AF_INET;
    sockaddr.sin_port = htons((short) (port & 0xFFFF));
    hostent = gethostbyname(host);
    if (hostent != NULL) {
	memcpy((char *) &addr,
	    (char *) hostent->h_addr_list[0], (size_t) hostent->h_length);
    } else {
	addr.s_addr = inet_addr(host);
	if (addr.s_addr == (unsigned long)-1) {
	    goto error;
	}
    }

    /*
     * There is a rumor that this assignment may require care on
     * some 64 bit machines.
     */

    sockaddr.sin_addr.s_addr = addr.s_addr;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock == INVALID_SOCKET) {
	goto error;
    }

    /*
     * Set kernel space buffering
     */

    TclSockMinimumBuffers(sock, TCP_BUFFER_SIZE);

    /*
     * Make socket close down gracefully.
     */
    
    ling.l_onoff = 1;
    ling.l_linger = 0;
    (void) setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *) &ling,
            sizeof(ling));

    if (server) {
	/*
	 * Set up to reuse server addresses automatically and bind to the
	 * specified port.
	 */
    
	status = 1;
	(void) setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &status,
		sizeof(status));
	status = bind(sock, (struct sockaddr *) &sockaddr, sizeof(sockaddr));
	if (status != SOCKET_ERROR) {
	    listen(sock, 5);
	}
    } else {
	/*
	 * Attempt to connect. The connect may fail at present with an
	 * EINPROGRESS but at a later time it will complete. The caller
	 * will set up a file handler on the socket if she is interested in
	 * being informed when the connect completes.
	 */

	status = connect(sock, (struct sockaddr *) &sockaddr,
		sizeof(sockaddr));
    }

    if (status == SOCKET_ERROR) {
	TclWinConvertWSAError(WSAGetLastError());
        if (interp != NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        closesocket(sock);
        return NULL;
    }

    /*
     * Allocate a new SocketInfo for this socket.
     */

    infoPtr = (SocketInfo *) ckalloc((unsigned) sizeof(SocketInfo));
    infoPtr->flags = 0;
    infoPtr->checkMask = 0;
    infoPtr->watchMask = 0;
    infoPtr->socket = sock;
    infoPtr->file = Tcl_GetFile((ClientData) infoPtr, TCL_WIN_SOCKET);

    /*
     * Add this socket to the global list of sockets.
     */

    infoPtr->nextPtr = socketList;
    socketList = infoPtr;

    return infoPtr;

  error:
    TclWinConvertWSAError(WSAGetLastError());
    if (interp != NULL) {
	Tcl_AppendResult(interp, "couldn't open socket: ",
		Tcl_PosixError(interp), (char *) NULL);
    }
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpClient --
 *
 *	Opens a TCP client socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed.  An error message is returned
 *	in the interpreter on failure.
 *
 * Side effects:
 *	Opens a client socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpClient(interp, port, host)
    Tcl_Interp *interp;			/* For error reporting; can be NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Host on which to open port. */
{
    Tcl_Channel chan;
    SocketInfo *infoPtr;
    char channelName[20];

    if (TclHasSockets(interp) != TCL_OK) {
	return NULL;
    }

    /*
     * Create a new client socket and wrap it in a channel.
     */

    infoPtr = CreateSocket(interp, port, host, 0);
    if (infoPtr == NULL) {
	return NULL;
    }

    infoPtr->acceptProc = NULL;
    infoPtr->acceptProcData = (ClientData) NULL;

    sprintf(channelName, "sock%d", infoPtr->socket);

    chan = Tcl_CreateChannel(&tcpChannelType, channelName, infoPtr->file,
	    infoPtr->file, (ClientData) infoPtr);

    /*
     * Sockets start out with EOL translation set to "auto".
     */

    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpServer --
 *
 *	Opens a TCP server socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed.  An error message is returned
 *	in the interpreter on failure.
 *
 * Side effects:
 *	Opens a server socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpServer(interp, port, host, acceptProc, acceptProcData)
    Tcl_Interp *interp;			/* For error reporting - may be
                                         * NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Name of local host. */
    Tcl_TcpAcceptProc *acceptProc;	/* Callback for accepting connections
                                         * from new clients. */
    ClientData acceptProcData;		/* Data for the callback. */
{
    Tcl_Channel chan;
    SocketInfo *infoPtr;
    char channelName[20];

    if (TclHasSockets(interp) != TCL_OK) {
	return NULL;
    }

    /*
     * Create a new client socket and wrap it in a channel.
     */

    infoPtr = CreateSocket(interp, port, host, 1);
    if (infoPtr == NULL) {
	return NULL;
    }

    infoPtr->acceptProc = acceptProc;
    infoPtr->acceptProcData = acceptProcData;

    /*
     * Set up the callback mechanism for accepting connections
     * from new clients. The caller will use Tcl_TcpRegisterCallback
     * to register a callback to call when a new connection is
     * accepted.
     */

    Tcl_CreateFileHandler(infoPtr->file, TCL_READABLE, TcpAccept,
            (ClientData) infoPtr);

    sprintf(channelName, "sock%d", infoPtr->socket);

    chan = Tcl_CreateChannel(&tcpChannelType, channelName, NULL, NULL,
            (ClientData) infoPtr);

    /*
     * Sockets start out with EOL translation set to "auto".
     */

    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpAccept --
 *	Accept a TCP socket connection.  This is called by the event loop,
 *	and it in turns calls any registered callbacks for this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Evals the Tcl script associated with the server socket.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
TcpAccept(data, mask)
    ClientData data;			/* Callback token. */
    int mask;				/* Not used. */
{
    SocketInfo *infoPtr = (SocketInfo *) data;
    SocketInfo *newInfoPtr;
    SOCKET sock;
    struct sockaddr_in addr;		/* The remote address */
    int len;				/* For accept interface */
    Tcl_Channel  chan;			/* Channel instance created. */
    char channelName[20];

    len = sizeof(struct sockaddr_in);
    sock = accept(infoPtr->socket, (struct sockaddr *)&addr, &len);

    infoPtr->checkMask &= ~TCL_READABLE;

    if (sock == INVALID_SOCKET) {
        return;
    }

    /*
     * Clear the inherited event mask.
     */

    WSAAsyncSelect(sock, socketWindow, 0, 0);

    newInfoPtr = (SocketInfo *) ckalloc((unsigned) sizeof(SocketInfo));
    newInfoPtr->flags = 0;
    newInfoPtr->checkMask = 0;
    newInfoPtr->watchMask = 0;
    newInfoPtr->socket = sock;
    newInfoPtr->acceptProc = NULL;
    newInfoPtr->acceptProcData = (ClientData) NULL;
    newInfoPtr->file = Tcl_GetFile((ClientData) newInfoPtr, TCL_WIN_SOCKET);

    sprintf(channelName, "sock%d", newInfoPtr->socket);

    chan = Tcl_CreateChannel(&tcpChannelType, channelName, newInfoPtr->file,
	    newInfoPtr->file, (ClientData) newInfoPtr);

    if (chan == (Tcl_Channel) NULL) {
	closesocket(newInfoPtr->socket);
	Tcl_FreeFile(newInfoPtr->file);
	ckfree((char *) newInfoPtr);
	return;
    } 

    /*
     * Add this socket to the global list of sockets.
     */

    newInfoPtr->nextPtr = socketList;
    socketList = newInfoPtr;

    /*
     * Invoke the accept callback procedure.
     */

    if (infoPtr->acceptProc != NULL) {
	(infoPtr->acceptProc) (infoPtr->acceptProcData, chan,
		inet_ntoa(addr.sin_addr));
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TcpInput --
 *
 *	Reads from a TCP socket under Windows.
 *
 * Results:
 *	The number of bytes read or -1 on error.
 *
 * Side effects:
 *	Consumes input from the socket.
 *
 *----------------------------------------------------------------------
 */

static int
TcpInput(instanceData, inFile, buf, toRead, errorCodePtr)
    ClientData instanceData;		/* The socket state. */
    Tcl_File inFile;			/* Not used. */
    char *buf;				/* Where to store data. */
    int toRead;				/* Maximum number of bytes to read. */
    int *errorCodePtr;			/* Where to store error codes. */
{
    SocketInfo *infoPtr = (SocketInfo *) instanceData;
    unsigned long bytesAvailable;
    int bytesRead;
    
    *errorCodePtr = 0;

    if (ioctlsocket(infoPtr->socket, FIONREAD, &bytesAvailable) == 0) {
        if (bytesAvailable < toRead) {
            toRead = bytesAvailable;
        }
        if ((toRead == 0) && !(infoPtr->flags & SOCKET_ASYNCH)) {
            toRead = 1;
        }
    } else {
        TclWinConvertWSAError(WSAGetLastError());
        if (errno == ECONNRESET) {
            return 0;
        }
        *errorCodePtr = errno;
        return -1;
    }

    if ((toRead == 0) && (bytesAvailable == 0)) {
        *errorCodePtr = EAGAIN;
        return -1;
    }
    bytesRead = recv(infoPtr->socket, buf, toRead, 0);
    if (bytesRead == SOCKET_ERROR) {
	TclWinConvertWSAError(WSAGetLastError());
        if (errno == ECONNRESET) {
            return 0;
        }
        *errorCodePtr = errno;
        return -1;
    }

    /*
     * Clear the readable bit in the check mask.  If an async handler
     * is still registered for this socket, then it will generate a new
     * event if there is still data available.  When the event is
     * processed, the readable bit will be turned back on.
     */

    infoPtr->checkMask &= ~TCL_READABLE;

    return bytesRead;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpOutput --
 *
 *	Writes data to a socket under Windows.
 *
 * Results:
 *	The number of bytes written or -1 on failure.
 *
 * Side effects:
 *	Produces output on the socket.
 *
 *----------------------------------------------------------------------
 */

static int
TcpOutput(instanceData, outFile, buf, toWrite, errorCodePtr)
    ClientData instanceData;		/* The socket state. */
    Tcl_File outFile;			/* The socket to write to. */
    char *buf;				/* Where to get data. */
    int toWrite;			/* Maximum number of bytes to write. */
    int *errorCodePtr;			/* Where to store error codes. */
{
    SocketInfo *infoPtr = (SocketInfo *) instanceData;
    int bytesWritten;

    *errorCodePtr = 0;

    bytesWritten = send(infoPtr->socket, buf, toWrite, 0);
    if (bytesWritten == SOCKET_ERROR) {
	TclWinConvertWSAError(WSAGetLastError());
        *errorCodePtr = errno;
        return -1;
    }

    /*
     * Clear the readable bit in the check mask.  If an async handler
     * is still registered for this socket, then it will generate a new
     * event if there is still data available.  When the event is
     * processed, the readable bit will be turned back on.
     */

    infoPtr->checkMask &= ~TCL_WRITABLE;

    return bytesWritten;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpBlockMode--
 *
 *	Implements blocking and nonblocking on a socket under Windows.
 *
 * Results:
 *	0 on success, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or nonblocking mode
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpBlockMode(instanceData, inFile, outFile, mode)
    ClientData instanceData;		/* Socket state. */
    Tcl_File inFile;			/* Input side of socket. */
    Tcl_File outFile;			/* Output side of socket. */
    int mode;				/* The mode to set: one of
                                         * TCL_MODE_BLOCKING or
                                         * TCL_MODE_NONBLOCKING. */
{
    SocketInfo *infoPtr = (SocketInfo *) instanceData;

    if (mode == TCL_MODE_BLOCKING) {
	infoPtr->flags |= SOCKET_ASYNCH;
    } else {
	infoPtr->flags &= ~SOCKET_ASYNCH;
    }
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpBufferSize --
 *
 *	Returns the preferred buffer size for TCP channels.
 *
 * Results:
 *	The preferred buffer size.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpBufferSize(clientData)
    ClientData clientData;		/* Unused. */
{
    return TCP_BUFFER_SIZE;
}

/*
 *----------------------------------------------------------------------
 *
 * TclWinWatchSocket --
 *
 *	This function imlements the socket specific portion of the
 *	Tcl_WatchFile function in the notifier.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The watched socket will be placed into non-blocking mode, and
 *	an entry on the asynch handler list will be created if necessary. 
 *
 *----------------------------------------------------------------------
 */

void
TclWinWatchSocket(file, mask)
    Tcl_File file;		/* Socket to watch. */
    int mask;			/* OR'ed combination of TCL_READABLE,
				 * TCL_WRITABLE, and TCL_EXCEPTION:
				 * indicates conditions to wait for
				 * in select. */
{
    SocketInfo *infoPtr = (SocketInfo *) Tcl_GetFileInfo(file, NULL);

    infoPtr->flags |= SOCKET_WATCH;

    /*
     * If the new mask includes more conditions than the current mask,
     * then we mark the socket as unregistered so it will be reregistered
     * the next time we enter Tcl_WaitForEvent.
     */

    mask |= infoPtr->watchMask;
    if (infoPtr->watchMask != mask) {
	infoPtr->flags &= ~SOCKET_REGISTERED;
	infoPtr->watchMask = mask;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TclWinNotifySocket --
 *
 *	Set up event notifiers for any sockets that are being watched.
 *	Also, clean up any sockets that are no longer being watched.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Adds and removes asynch select handlers.
 *
 *----------------------------------------------------------------------
 */

void
TclWinNotifySocket()
{
    SocketInfo *infoPtr;

    if (socketList == NULL) {
	return;
    }

    /*
     * Establish or remove any notifiers.
     */

    for (infoPtr = socketList; infoPtr != NULL; infoPtr = infoPtr->nextPtr) {
	if (infoPtr->flags & SOCKET_WATCH) {
	    if (!(infoPtr->flags & SOCKET_REGISTERED)) {
		int events = 0;

		if (infoPtr->watchMask & TCL_READABLE) {
		    events |= (FD_READ | FD_ACCEPT | FD_CLOSE);
		} else if (infoPtr->watchMask & TCL_WRITABLE) {
		    events |= (FD_WRITE | FD_CONNECT);
		}
		WSAAsyncSelect(infoPtr->socket, socketWindow, SOCKET_MESSAGE,
			events);
	    }
	} else {
	    if (infoPtr->flags & SOCKET_REGISTERED) {
		WSAAsyncSelect(infoPtr->socket, socketWindow, 0, 0);
	    }
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TclWinSocketReady --
 *
 *	This function is invoked by Tcl_FileReady to check whether
 *	the specified conditions are present on a socket.
 *
 * Results:
 *	The return value is 0 if none of the conditions specified by
 *	mask were true for socket the last time the system checked.
 *	If any of the conditions were true, then the return value is a
 *	mask of those that were true.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
TclWinSocketReady(file, mask)
    Tcl_File file;	/* File handle for a stream. */
    int mask;			/* OR'ed combination of TCL_READABLE,
				 * TCL_WRITABLE, and TCL_EXCEPTION:
				 * indicates conditions caller cares about. */
{
    SocketInfo *infoPtr = (SocketInfo *) Tcl_GetFileInfo(file, NULL);

    return (infoPtr->checkMask & mask);
}

/*
 *----------------------------------------------------------------------
 *
 * SocketProc --
 *
 *	This function is called when WSAAsyncSelect has been used
 *	to register interest in a socket event, and the event has
 *	occurred.
 *
 * Results:
 *	0 on success.
 *
 * Side effects:
 *	The checkMask for the given socket is updated to reflect the
 *	event which occured.
 *
 *----------------------------------------------------------------------
 */

static LRESULT CALLBACK
SocketProc(hwnd, message, wParam, lParam)
    HWND hwnd;
    UINT message;
    WPARAM wParam;
    LPARAM lParam;
{
    int event;
    SOCKET socket;
    SocketInfo *infoPtr;

    if ((hwnd != socketWindow) || (message != SOCKET_MESSAGE)) {
	return DefWindowProc(hwnd, message, wParam, lParam);
    }
    event = WSAGETSELECTEVENT(lParam);
    socket = (SOCKET) wParam;

    /*
     * Find the specified socket on the socket list and update its
     * check flags.
     */

    for (infoPtr = socketList; infoPtr != NULL; infoPtr = infoPtr->nextPtr) {
	if (infoPtr->socket == socket) {

	    if (event & (FD_READ | FD_ACCEPT | FD_CLOSE)) {
		infoPtr->checkMask |= TCL_READABLE;
	    }
	    if (event & (FD_WRITE | FD_CONNECT)) {
		infoPtr->checkMask |= TCL_WRITABLE;
	    }
	    break;
	}
    }
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_GetHostName --
 *
 *	Returns the name of the local host.
 *
 * Results:
 *	Returns a string containing the host name, or NULL on error.
 *	The returned string must be freed by the caller.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

char *
Tcl_GetHostName()
{
    static int  hostnameInitialized = 0;
    static char hostname[255];	/* This buffer should be big enough for
                                 * hostname plus domain name. */

    if (TclHasSockets(NULL) != TCL_OK) {
	return "";
    }

    if (hostnameInitialized) {
        return hostname;
    }
    if (gethostname(hostname, 100) == 0) {
        hostnameInitialized = 1;
        return hostname;
    }
    return (char *) NULL;
}
